/**** <lalVerbatim file="FrameSeriesCV">
 * Author: Jolien D. E. Creighton
 * $Id$
 **** </lalVerbatim> */

/**** <lalLaTeX>
 *
 * \subsection{Module \texttt{FrameSeries.c}}
 *
 * These routines are used to read/write time series data from/to frame files.
 *
 * \subsubsection*{Prototypes}
 * \input{FrameSeriesCP}
 * \idx{LALFrGetINT2TimeSeries}
 * \idx{LALFrGetINT2TimeSeriesMetadata}
 * \idx{LALFrGetINT4TimeSeries}
 * \idx{LALFrGetINT4TimeSeriesMetadata}
 * \idx{LALFrGetINT8TimeSeries}
 * \idx{LALFrGetINT8TimeSeriesMetadata}
 * \idx{LALFrGetREAL4TimeSeries}
 * \idx{LALFrGetREAL4TimeSeriesMetadata}
 * \idx{LALFrGetREAL8TimeSeries}
 * \idx{LALFrGetREAL8TimeSeriesMetadata}
 * \idx{LALFrGetCOMPLEX8TimeSeries}
 * \idx{LALFrGetCOMPLEX8TimeSeriesMetadata}
 * \idx{LALFrGetCOMPLEX16TimeSeries}
 * \idx{LALFrGetCOMPLEX16TimeSeriesMetadata}
 * \idx{LALFrGetINT2FrequencySeries}
 * \idx{LALFrGetINT2FrequencySeriesMetadata}
 * \idx{LALFrGetINT4FrequencySeries}
 * \idx{LALFrGetINT4FrequencySeriesMetadata}
 * \idx{LALFrGetINT8FrequencySeries}
 * \idx{LALFrGetINT8FrequencySeriesMetadata}
 * \idx{LALFrGetREAL4FrequencySeries}
 * \idx{LALFrGetREAL4FrequencySeriesMetadata}
 * \idx{LALFrGetREAL8FrequencySeries}
 * \idx{LALFrGetREAL8FrequencySeriesMetadata}
 * \idx{LALFrGetCOMPLEX8FrequencySeries}
 * \idx{LALFrGetCOMPLEX8FrequencySeriesMetadata}
 * \idx{LALFrGetCOMPLEX16FrequencySeries}
 * \idx{LALFrGetCOMPLEX16FrequencySeriesMetadata}
 * \idx{LALFrWriteINT2TimeSeries}
 * \idx{LALFrWriteINT4TimeSeries}
 * \idx{LALFrWriteINT8TimeSeries}
 * \idx{LALFrWriteREAL4TimeSeries}
 * \idx{LALFrWriteREAL8TimeSeries}
 * \idx{LALFrWriteCOMPLEX8TimeSeries}
 * \idx{LALFrWriteCOMPLEX16TimeSeries}
 * \idx{LALFrGetTimeSeriesType}
 *
 * \subsubsection*{Description}
 *
 * The routines
 * \texttt{LALFrGet}$\langle\mbox{datatype}\rangle$\texttt{TimeSeries()}
 * search the frame for a specified channel.  If the time series supplied has
 * data storage allocated, then the specified amount of data is filled from
 * the frame stream.  If no space has been allocated (so that the data field
 * is \texttt{NULL}), then only the channel information is returned in the
 * time series (e.g., the start time of the next data and the time step size).
 *
 * Because it is good coding practice to not tinker directly with the innards
 * of structures like time series, the behaviour described above is undesirable
 * whereby the time series reading functions do useful things when part of the
 * time series structures passed into them is not initialized.  To address
 * this, the routines
 * \texttt{LALFrGet}$\langle\mbox{datatype}\rangle$\texttt{TimeSeriesMetadata()}
 * are provided.  These routines accept a fully initialized time series
 * structure, and poplulate only the metadata from the frame stream.  New code
 * should be written to use \emph{these} functions when only the time series
 * meta data is desired.
 *
 * The routine \texttt{LALFrGetTimeSeriesType} returns the type of the time
 * series corresponding to the specified channel.  This is needed if it is not
 * known in advance what type of data is stored in the channel.
 *
 * The routines
 * \texttt{LALFrWrite}$\langle\mbox{datatype}\rangle$\texttt{TimeSeries()}
 * outputs a given time series as a new frame file with the filename
 * $\langle\mbox{source}\rangle$\verb+-+$\langle\mbox{description}\rangle$%
 * \verb+-+$\langle\mbox{GPS-seconds}\rangle$\verb+-+%
 * $\langle\mbox{duration}\rangle$\verb+.gwf+
 * where source and description are the specified frame source and description
 * identifiers, GPS-seconds is the start time of the data in
 * seconds since 0h UTC 6 Jan 1980 (GPS time origin), or the nearest second
 * before the start of the data, and duration is the number of seconds between
 * value of GPS-seconds and the GPS end time of the data, rounded up.
 *
 * \vfill{\footnotesize\input{FrameSeriesCV}}
 *
 **** </lalLaTeX> */


int rename( const char *from, const char *to );


#include <lal/LALFrameIO.h>

#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <lal/LALStdio.h>
#include <lal/LALStdlib.h>
#include <lal/Date.h>
#include <lal/Units.h>
#include <lal/FrameCache.h>
#include <lal/FrameStream.h>
#include <lal/TimeSeries.h>
#include <lal/LALDetectors.h>
#include <lal/LALDatatypes.h>

NRCSID( FRAMESERIESC, "$Id$" );

/* Useful macros */
#define SECNAN_TO_I8TIME( sec, nan ) \
  ((INT8)1000000000*(INT8)(sec)+(INT8)(nan))
/* Dangerous!!!: */
#define EPOCH_TO_I8TIME( epoch ) \
  SECNAN_TO_I8TIME( (epoch).gpsSeconds, (epoch).gpsNanoSeconds )
#define SET_EPOCH( pepoch, i8time ) \
  do { INT8 t=(i8time); LIGOTimeGPS *pe=(pepoch); \
    pe->gpsSeconds=t/(INT8)1000000000; pe->gpsNanoSeconds=t%(INT8)1000000000; \
  } while( 0 )

#ifdef __GNUC__
#define UNUSED __attribute__ ((unused))
#else
#define UNUSED
#endif

/*
 *
 * Routine to load a FrVect associated with a given channel from a given frame.
 *
 */
static struct FrVect *loadFrVect( FrStream *stream, const char *channel )
{
  char        chan[256];
  FrAdcData  *adc;
  FrSimData  *sim;
  FrProcData *proc;
  FrChanType  type;
  FrVect     *vect;
  FrTOCts    *ts;
  double      tstart;
  strncpy( chan, channel, sizeof( chan ) - 1 );
  if ( ! stream->file->toc )
  {
    if ( FrTOCReadFull( stream->file ) == 0 )
    {
      stream->state |= LAL_FR_ERR | LAL_FR_TOC;
      return NULL;
    }
  }
  stream->file->relocation = FR_NO;
  tstart  = stream->file->toc->GTimeS[stream->pos];
  tstart += 1e-9 * stream->file->toc->GTimeN[stream->pos];
  ts = stream->file->toc->adc;
  /* scan adc data channels */
  type = LAL_ADC_CHAN;
  while ( ts && strcmp( channel, ts->name ) )
    ts = ts->next;
  if ( ! ts )
  {
    /* scan sim data channels */
    type = LAL_SIM_CHAN;
    ts = stream->file->toc->sim;
    while ( ts && strcmp( channel, ts->name ) )
      ts = ts->next;
  }
  if ( ! ts )
  {
    /* scan proc data channels */
    type = LAL_PROC_CHAN;
    ts = stream->file->toc->proc;
    while ( ts && strcmp( channel, ts->name ) )
      ts = ts->next;
  }
  if ( ! ts )
    return NULL;

  FrTOCSetPos( stream->file, ts->position[stream->pos] );
  switch ( type )
  {
    case LAL_ADC_CHAN:
      adc = FrAdcDataRead( stream->file );
      if ( ! adc ) /* only happens if memory allocation error */
        return NULL;
      adc->next = NULL;
      vect = FrVectReadNext( stream->file, tstart, chan );
      FrAdcDataFree( adc );
      break;
    case LAL_SIM_CHAN:
      sim = FrSimDataRead( stream->file );
      if ( ! sim ) /* only happens if memory allocation error */
        return NULL;
      sim->next = NULL;
      vect = FrVectReadNext( stream->file, tstart, chan );
      FrSimDataFree( sim );
      break;
    case LAL_PROC_CHAN:
      proc = FrProcDataRead( stream->file );
      if ( ! proc ) /* only happens if memory allocation error */
        return NULL;
      proc->next = NULL;
      vect = FrVectReadNext( stream->file, tstart, chan );
      FrProcDataFree( proc );
      break;
    default:
      return NULL;
  }
  if ( vect && vect->compress )
    FrVectExpand( vect );
  return vect;
}


/*
 *
 * Routine to make a 1D FrVect structure and attach it to a frame as the
 * appropriate channel type.
 *
 */
static struct FrVect *makeFrVect1D( struct FrameH *frame, int chtype,
    char *name, char *comment, char *unitx, char *unity, int datatype,
    double rate, double fshift, double dx, unsigned int npts )
{
  struct FrVect *vect = FrVectNew1D( name, datatype, npts, dx, unitx, unity );
  if ( ! vect ) return NULL;
  if ( chtype == ProcDataChannel )
  {
    struct FrProcData *proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      FrVectFree( vect );
      return NULL;
    }
    proc->classe = FrProcDataDef();
    proc->fShift = fshift;
    proc->type = 1; /* time series */
    proc->data = vect;
    proc->next = frame->procData;
    frame->procData = proc;
    if ( ! FrStrCpy( &proc->name, name ) ) return NULL;
    if ( ! FrStrCpy( &proc->comment, comment ) ) return NULL;
  }
  else if ( chtype == ADCDataChannel )
  {
    struct FrAdcData *adc = calloc( 1, sizeof( *adc ) );
    struct FrRawData *raw = frame->rawData ? frame->rawData :
      FrRawDataNew( frame );
    if ( ! adc || ! raw )
    {
      FrVectFree( vect );
      if ( adc ) free( adc );
      return NULL;
    }
    adc->classe = FrAdcDataDef();
    adc->sampleRate = rate;
    adc->fShift = fshift;
    adc->data = vect;
    adc->next = raw->firstAdc;
    raw->firstAdc = adc;
    if ( ! FrStrCpy( &adc->name, name ) ) return NULL;
    if ( ! FrStrCpy( &adc->comment, comment ) ) return NULL;
    if ( ! FrStrCpy( &adc->units, unity ) ) return NULL;
  }
  else if ( chtype == SimDataChannel )
  {
    struct FrSimData *sim = calloc( 1, sizeof( *sim ) );
    if ( ! sim )
    {
      FrVectFree( vect );
      return NULL;
    }
    sim->classe = FrSimDataDef();
    sim->sampleRate = rate;
    sim->data = vect;
    sim->next = frame->simData;
    frame->simData = sim;
    if ( ! FrStrCpy( &sim->name, name ) ) return NULL;
    if ( ! FrStrCpy( &sim->comment, comment ) ) return NULL;
  }
  else
  {
    FrVectFree( vect );
    return NULL;
  }
  return vect;
}

static FrVect * FrVectReadInfo( FrFile *iFile, FRULONG *pos )
{
  FrVect *v;
  unsigned short type;
  FRULONG localpos;
  if ( FrFileIGoToNextRecord( iFile ) != iFile->vectorType )
    return NULL;
  v = calloc( 1, sizeof( FrVect ) );
  if ( ! v )
  {
    iFile->error = FR_ERROR_MALLOC_FAILED;
    return NULL;
  }
  FrReadHeader( iFile, v );
  FrReadSChar( iFile, &v->name );
  FrReadShortU( iFile, &v->compress );
  if ( v->compress == 256 )
    v->compress = 0; /* we will swap bytes at reading time */
  FrReadShortU( iFile, &type );
  v->type = type;
  switch ( v->type )
  {
    case FR_VECT_4R:
      v->wSize = sizeof( float );
      break;
    case FR_VECT_8R:
      v->wSize = sizeof( double );
      break;
    case FR_VECT_C:
      v->wSize = sizeof( char );
      break;
    case FR_VECT_1U:
      v->wSize = sizeof( char );
      break;
    case FR_VECT_2S:
      v->wSize = sizeof( short );
      break;
    case FR_VECT_2U:
      v->wSize = sizeof( short );
      break;
    case FR_VECT_4S:
      v->wSize = sizeof( int );
      break;
    case FR_VECT_4U:
      v->wSize = sizeof( int );
      break;
    case FR_VECT_8S:
      v->wSize = sizeof( FRLONG );
      break;
    case FR_VECT_8U:
      v->wSize = sizeof( FRLONG );
      break;
    case FR_VECT_8C:
      v->wSize = 2 * sizeof( float );
      break;
    case FR_VECT_16C:
      v->wSize = 2 * sizeof( double );
      break;
    case FR_VECT_8H:
      v->wSize = 2 * sizeof( float );
      break;
    case FR_VECT_16H:
      v->wSize = 2 * sizeof( double );
      break;
    default:
      v->wSize = 0;
  }
  if ( iFile->fmtVersion > 5 )
  {
    FrReadLong( iFile, (FRLONG *)&v->nData );
    FrReadLong( iFile, (FRLONG *)&v->nBytes );
  }
  else
  {
    unsigned int nData, nBytes;
    FrReadIntU( iFile, &nData );
    FrReadIntU( iFile, &nBytes );
    v->nData  = nData;
    v->nBytes = nBytes;
  }
  v->space = v->nData;

  /* skip the data */
  localpos = FrIOTell( iFile->frfd );
  if ( pos )
    *pos = localpos;
  localpos += v->nBytes;
  FrIOSet( iFile->frfd, localpos );
  FrReadIntU( iFile, &v->nDim );
  FrReadVL( iFile, (FRLONG**)&v->nx, v->nDim );
  FrReadVD( iFile, &v->dx, v->nDim );
  FrReadVD( iFile, &v->startX, v->nDim );
  FrReadVQ( iFile, &v->unitX, v->nDim );
  FrReadSChar( iFile, &v->unitY );
  FrReadStruct( iFile, &v->next );

  if ( pos )
    *pos = localpos;
  return v;
}

LALTYPECODE XLALFrGetTimeSeriesType( const char *channel, FrStream *stream )
{
  static const char func[] = "XLALFrGetTimeSeriesType";
  FrChanType chantype;
  FrTOCts    *ts   = NULL;
  FrProcData *proc = NULL;
  FrAdcData  *adc  = NULL;
  FrSimData  *sim  = NULL;
  int type = -1;

  if ( ! channel || ! stream )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  if ( ! stream->file->toc )
  {
    if ( FrTOCReadFull( stream->file ) == NULL )
    {
      XLALPrintError( "XLAL Error - %s: could not open frame TOC %s\n",
          func, stream->file );
      stream->state |= LAL_FR_ERR | LAL_FR_TOC;
      XLAL_ERROR( func, XLAL_EIO );
    }
  }

  /* scan adc channels */
  chantype = LAL_ADC_CHAN;
  ts = stream->file->toc->adc;
  while ( ts && strcmp( channel, ts->name ) )
    ts = ts->next;

  if ( ! ts )
  {
    /* scan sim channels */
    chantype = LAL_SIM_CHAN;
    ts = stream->file->toc->sim;
    while ( ts && strcmp( channel, ts->name ) )
      ts = ts->next;
  }

  if ( ! ts )
  {
    /* scan proc channels */
    chantype = LAL_PROC_CHAN;
    ts = stream->file->toc->proc;
    while ( ts && strcmp( channel, ts->name ) )
      ts = ts->next;
  }

  if ( ts ) /* the channel was found */
  {
    FrTOCSetPos( stream->file, ts->position[0] );
    switch ( chantype )
    {
      case LAL_ADC_CHAN:
        if ( ( adc = FrAdcDataRead( stream->file ) ) )
        {
          adc->data = FrVectReadInfo( stream->file, NULL );
          type = adc->data ? adc->data->type : -1;
          FrAdcDataFree( adc );
        }
        break;
      case LAL_SIM_CHAN:
        if ( ( sim = FrSimDataRead( stream->file ) ) )
        {
          sim->data = FrVectReadInfo( stream->file, NULL );
          type = sim->data ? sim->data->type : -1;
          FrSimDataFree( sim );
        }
        break;
      case LAL_PROC_CHAN:
        if ( ( proc = FrProcDataRead( stream->file ) ) )
        {
          proc->data = FrVectReadInfo( stream->file, NULL );
          type = proc->data ? proc->data->type : -1;
          FrProcDataFree( proc );
        }
        break;
      default:
        type = -1;
    }
  }

  switch ( type )
  {
    case FR_VECT_C:
      return LAL_CHAR_TYPE_CODE;
    case FR_VECT_2S:
      return LAL_I2_TYPE_CODE;
    case FR_VECT_4S:
      return LAL_I4_TYPE_CODE;
    case FR_VECT_8S:
      return LAL_I8_TYPE_CODE;
    case FR_VECT_1U:
      return LAL_UCHAR_TYPE_CODE;
    case FR_VECT_2U:
      return LAL_U2_TYPE_CODE;
    case FR_VECT_4U:
      return LAL_U4_TYPE_CODE;
    case FR_VECT_8U:
      return LAL_U8_TYPE_CODE;
    case FR_VECT_4R:
      return LAL_S_TYPE_CODE;
    case FR_VECT_8R:
      return LAL_D_TYPE_CODE;
    case FR_VECT_C8:
      return LAL_C_TYPE_CODE;
    case FR_VECT_C16:
      return LAL_Z_TYPE_CODE;
    default:
      XLAL_ERROR( func, XLAL_ETYPE );
  }

  XLAL_ERROR( func, XLAL_ETYPE );
}



/* little helper function for getting number of points in a channel */
int XLALFrGetVectorLength ( CHAR *name, FrStream *stream )
{
  static const char func[] = "XLALFrGetVectorLength";
  struct FrVect	*vect;
  int ret = -1;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */

  ret = vect->nData;

  FrVectFree(vect);

  return ret;

}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetTimeSeriesType(
    LALStatus   *status,
    LALTYPECODE *output,
    FrChanIn    *chanin,
    FrStream    *stream
    )
{ /* </lalVerbatim> */
  int type;
  INITSTATUS( status, "LALFrGetTimeSeriesType", FRAMESERIESC );

  ASSERT( output, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( chanin, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( chanin->name, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  type = XLALFrGetTimeSeriesType( chanin->name, stream );
  if ( type < 0 )
  {
    int errnum = xlalErrno;
    XLALClearErrno();
    switch ( errnum )
    {
      case XLAL_EIO:
        ABORT( status, FRAMESTREAMH_EREAD, FRAMESTREAMH_MSGEREAD );
      case XLAL_ETYPE:
        ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
      default:
        ABORTXLAL( status );
    }
  }

  *output = type;
  RETURN( status );
}


static int copy_FrVect_to_REAL8( REAL8 *data, struct FrVect *vect, size_t ncpy, size_t offset )
{
	size_t i;
	if ( ncpy + offset > vect->nData )
		return -1;
	switch ( vect->type ) {
		case FR_VECT_C:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->data[offset + i];
			break;
		case FR_VECT_1U:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataU[offset + i];
			break;
		case FR_VECT_2U:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataUS[offset + i];
			break;
		case FR_VECT_4U:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataUI[offset + i];
			break;
		case FR_VECT_8U:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataUL[offset + i];
			break;
		case FR_VECT_2S:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataS[offset + i];
			break;
		case FR_VECT_4S:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataI[offset + i];
			break;
		case FR_VECT_8S:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataL[offset + i];
			break;
		case FR_VECT_4R:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataF[offset + i];
			break;
		case FR_VECT_8R:
			for ( i = 0; i < ncpy; ++i )
			       	data[i] = vect->dataD[offset + i];
			break;
		default:
			return -1;
	}
	return ncpy;
}

REAL8TimeSeries * XLALFrInputREAL8TimeSeries( FrStream *stream, const char *channel, const LIGOTimeGPS *start, REAL8 duration, size_t lengthlimit )
{
	static const char func[] = "XLALFrInputREAL8TimeSeries";
	const REAL8 fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
	struct FrVect *vect;
	REAL8TimeSeries *series;
	LIGOTimeGPS epoch;
	UINT4  need;
	UINT4  noff;
	UINT4  ncpy;
	REAL8 *dest;
	INT8   tnow;
	INT8   tbeg;
	INT8   tend;
	REAL8  rate;
	int    code;

	if ( stream->state & LAL_FR_ERR )
		XLAL_ERROR_NULL( func, XLAL_EIO );
	if ( stream->state & LAL_FR_END )
		XLAL_ERROR_NULL( func, XLAL_EIO );

	/* seek to the correct place */
	if ( XLALFrSeek( stream, start ) )
		XLAL_ERROR_NULL( func, XLAL_EIO );

	vect = loadFrVect( stream, channel );
	if ( ! vect || ! vect->data )
		XLAL_ERROR_NULL( func, XLAL_ENAME ); /* couldn't find channel */

	tnow = EPOCH_TO_I8TIME( stream->epoch );
#	if defined FR_VERS && FR_VERS >= 5000
	tbeg = 1e9 * vect->GTime;
#	else
	tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#	endif
	if ( tnow + 1000 < tbeg ) { /* added 1000 ns to account for double precision */
		FrVectFree(vect);
		XLAL_ERROR_NULL( func, XLAL_ETIME ); /* invalid time offset */
	}

	/* compute number of points offset very carefully:
	 * if current time is within fuzz of a sample, get that sample;
	 * otherwise get the sample just after the requested time */
	rate = 1.0 / vect->dx[0];
	noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );
	if ( noff > vect->nData ) {
		FrVectFree(vect);
		XLAL_ERROR_NULL( func, XLAL_ETIME ); /* invalid time offset */
	}
	need = duration * rate;
	if ( lengthlimit && (lengthlimit < need) )
		need = lengthlimit;

	/* adjust current time to be exactly the first sample
	 * (rounded to nearest nanosecond) */
	tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );
	SET_EPOCH( &epoch, tnow );

	series = XLALCreateREAL8TimeSeries( channel, &epoch, 0.0, vect->dx[0], &lalADCCountUnit, need );
	if ( ! series )
		XLAL_ERROR_NULL( func, XLAL_EFUNC );
	dest = series->data->data;

	/* number of points to copy */
	ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
	code = copy_FrVect_to_REAL8( dest, vect, ncpy, noff );
	if ( code < 0 ) { /* fails if vect has complex data type */
		if(vect) FrVectFree(vect);
		XLALDestroyREAL8TimeSeries( series );
		XLAL_ERROR_NULL( func, XLAL_ETYPE ); /* data has wrong type */
	}

	FrVectFree(vect);
	vect = NULL;

	dest += ncpy;
	need -= ncpy;

	/* if still data remaining */
	while ( need ) {
		if ( XLALFrNext( stream ) < 0 ) {
			if(vect) FrVectFree(vect);
			XLALDestroyREAL8TimeSeries( series );
			XLAL_ERROR_NULL( func, XLAL_EFUNC );
		}
		if ( stream->state & LAL_FR_END ) {
			if(vect) FrVectFree(vect);
			XLALDestroyREAL8TimeSeries( series );
			XLAL_ERROR_NULL( func, XLAL_EIO );
		}

		/* load more data */
		vect = loadFrVect( stream, series->name );
		if ( ! vect || ! vect->data ) { /* channel missing */
			XLALDestroyREAL8TimeSeries( series );
			XLAL_ERROR_NULL( func, XLAL_ENAME );
		}

		if ( stream->state & LAL_FR_GAP ) { /* failure: gap in data */
			if(vect) FrVectFree(vect);
			XLALDestroyREAL8TimeSeries( series );
			XLAL_ERROR_NULL( func, XLAL_ETIME );
		}

		/* copy data */
		ncpy = vect->nData < need ? vect->nData : need;
		code = copy_FrVect_to_REAL8( dest, vect, ncpy, 0 );
		if ( code < 0 ) { /* fails if vect has complex data type */
			if(vect) FrVectFree(vect);
			XLALDestroyREAL8TimeSeries( series );
			XLAL_ERROR_NULL( func, XLAL_ETYPE );
		}

		FrVectFree(vect);
		vect=NULL;

		dest += ncpy;
		need -= ncpy;
	}

	/* update stream start time very carefully:
	 * start time must be the exact time of the next sample, rounded to the
	 * nearest nanosecond */
	SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch ) + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

	/* check to see that we are still within current frame */
	tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
	tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
	if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) ) {
		int keepmode = stream->mode;
		LIGOTimeGPS keep;
		keep = stream->epoch;
		/* advance a frame -- failure is benign */
		stream->mode |= LAL_FR_IGNOREGAP_MODE;
		XLALFrNext( stream );
		if ( ! (stream->state & LAL_FR_GAP) )
			stream->epoch = keep;
		stream->mode = keepmode;
	}

	if ( stream->state & LAL_FR_ERR ) {
		XLALDestroyREAL8TimeSeries( series );
		XLAL_ERROR_NULL( func, XLAL_EIO );
	}

	return series;
}





























/* <lalVerbatim file="FrameSeriesCP"> */
int XLALFrGetCOMPLEX16FrequencySeries ( COMPLEX16FrequencySeries *series, FrStream *stream )
{ /* </lalVerbatim> */
  static const char func[] = "XLALFrGetCOMPLEX16FrequencySeries";
  struct FrVect	*vect;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_16C )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  memcpy( series->data->data, vect->dataD, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetCOMPLEX16FrequencySeries (
    LALStatus		*status,
    COMPLEX16FrequencySeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  struct FrVect	*vect;
  INITSTATUS( status, "LALFrGetCOMPLEX16FrequencySeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( ! series->data, status, FRAMESTREAMH_ENNUL, FRAMESTREAMH_MSGENNUL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_16C )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  memcpy( series->data->data, vect->dataD, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  vect=NULL;

  RETURN( status );
}


int XLALFrGetCOMPLEX16TimeSeriesMetadata ( COMPLEX16TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetCOMPLEX16TimeSeriesMetadata";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 noff;
  INT8		 tnow;
  INT8		 tbeg;
  REAL8          rate;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_16C )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  FrVectFree(vect);
  return 0;
}


int XLALFrGetCOMPLEX16TimeSeries ( COMPLEX16TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetCOMPLEX16TimeSeries";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  COMPLEX16 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_16C )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* end here... just metadata was requested */
  {
    FrVectFree(vect);
    return 0;
  }

  /* check to see if data vector is ok */
  if ( ! series->data->data )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data->length )
    XLAL_ERROR( func, XLAL_EBADLEN );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataD );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataD + noff * mult, ncpy*sizeof(*series->data->data) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    if ( XLALFrNext( stream ) < 0 )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EIO );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ENAME ); /* now channel is missing ... */
    }
    if ( vect->type != FR_VECT_16C )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ETYPE ); /* now type is wrong ... */
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataD, ncpy * sizeof( *series->data->data ) );


    FrVectFree(vect);
    vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos],
      stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    if ( XLALFrNext( stream ) < 0 ) {
      stream->mode = keepmode;
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( ! (stream->state & LAL_FR_GAP) )
      stream->epoch = keep;
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
    stream->state |= LAL_FR_GAP;
  /* FIXME: does this need to cause a failure if mode is set to fail on gaps? */

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetCOMPLEX16TimeSeries (
    LALStatus		*status,
    COMPLEX16TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  COMPLEX16 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  INITSTATUS( status, "LALFrGetCOMPLEX16TimeSeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_16C )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    FrVectFree(vect);
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataD );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataD + noff * mult, ncpy * sizeof( *series->data->data ) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;


  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_16C )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataD, ncpy * sizeof( *series->data->data ) );


  FrVectFree(vect);
  vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! (stream->state & LAL_FR_GAP) )
    {
      stream->epoch = keep;
    }
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
  {
    stream->state |= LAL_FR_GAP;
  }

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }

  /* remove this: the error will be reported on the *next* call! */
  /*
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }
  */

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetCOMPLEX16TimeSeriesMetadata (
    LALStatus		*status,
    COMPLEX16TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  void *sequence;

  INITSTATUS (status, "LALFrGetCOMPLEX16TimeSeriesMetadata", FRAMESERIESC);
  ATTATCHSTATUSPTR (status);

  ASSERT (series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);
  ASSERT (stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);

  /* save the sequence address, then wipe the series structure */
  sequence = series->data;
  memset (series, 0, sizeof(*series));

  /* call LALFrGetCOMPLEX16TimeSeries to populate the series' metadata */
  LALFrGetCOMPLEX16TimeSeries (status->statusPtr, series, chanin, stream);
  CHECKSTATUSPTR (status);

  /* restore the sequence address */
  series->data = sequence;

  DETATCHSTATUSPTR (status);
  RETURN (status);
}


COMPLEX16TimeSeries *XLALFrReadCOMPLEX16TimeSeries (
	FrStream *stream,
	const char *chname,
	const LIGOTimeGPS *start,
	REAL8 duration,
	size_t lengthlimit
)
{
	static const char func[] = "XLALFrReadCOMPLEX16TimeSeries";
	COMPLEX16TimeSeries *series;
	size_t length;

	/* create and initialize a zero-length time series vector */
	series = XLALCreateCOMPLEX16TimeSeries (chname, start, 0.0, 0.0, &lalADCCountUnit, 0);
	if(!series)
		XLAL_ERROR_NULL (func, XLAL_EFUNC);

	/* get the time series meta-data */
	if(XLALFrGetCOMPLEX16TimeSeriesMetadata (series, stream)) {
		XLALDestroyCOMPLEX16TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* resize the time series to the correct number of samples */
	length = duration / series->deltaT;
	if(lengthlimit && (lengthlimit < length))
		length = lengthlimit;
	if(!XLALResizeCOMPLEX16TimeSeries (series, 0, length)) {
		XLALDestroyCOMPLEX16TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* read the data */
	if(XLALFrSeek (stream, start) || XLALFrGetCOMPLEX16TimeSeries (series, stream)) {
		XLALDestroyCOMPLEX16TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	return(series);
}

























int XLALFrWriteCOMPLEX16TimeSeries ( COMPLEX16TimeSeries *series, int frnum )
{
  static const char func[] = "XLALFrWriteCOMPLEX16TimeSeries";
  char fname[FILENAME_MAX];
  char tmpfname[FILENAME_MAX];
  char comment[] = "XLALFrWriteCOMPLEX16TimeSeries $Id$";
  char seconds[] = "s";
  char units[LALUnitTextSize];
  const int run = 0; /* this routine always sets run number to zero */
  char site = 0;
  char ifo[3] = "\0\0\0";
  char description[FILENAME_MAX] = "";
  char *s;
  int detector;
  int detectorFlag = 0;
  REAL8 duration;
  INT4 t0;
  INT4 dt;
  struct FrProcData *proc;
  struct FrVect *vect;
  struct FrameH *frame;
  struct FrFile *frfile;
  int c;

  if ( ! series )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data || ! series->data->data )
    XLAL_ERROR( func, XLAL_EINVAL );

  /* if channel name is of the form Xn:... then Xn is the ifo name */
  if ( isupper( series->name[0] ) )
  {
    site = series->name[0];
    if ( isdigit( series->name[1] ) && series->name[2] == ':' )
    {
      memcpy( ifo, series->name, 2 );
      /* figure out what detector this is */
      for ( detector = 0; detector < LAL_NUM_DETECTORS; ++detector )
        if ( ! strcmp( ifo, lalCachedDetectors[detector].frDetector.prefix ) )
        {
          detectorFlag = 1 << 2 * detector;
          break;
        }
    }
  }

  /* if site was not set, do something sensible with it */
  if ( ! site )
    site = 'F'; /* fake site */

  /* transform channel name to be a valid description field of file name */
  memset( description, 0, sizeof( description ) );
  strncpy( description, series->name, sizeof( description) - 1 );
  s = description;
  while ( *s )
  {
    if ( ! isalnum( *s ) )
      *s = '_'; /* replace invalid description character with underscore */
    ++s;
  }

  /* compute duration of the data and the stop and end times of frame file */
  duration = series->data->length * series->deltaT;
  t0 = series->epoch.gpsSeconds;
  dt = (INT4)ceil( XLALGPSGetREAL8( &series->epoch ) + duration ) - t0;
  if ( t0 < 0 || dt < 1 )
    XLAL_ERROR( func, XLAL_ETIME );

  /* construct file name */
  c = snprintf( fname, sizeof( fname ), "%c-%s-%d-%d.gwf", site, description, t0, dt );
  if ( c < 0 || c > (int)sizeof(fname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );
  c = snprintf( tmpfname, sizeof( tmpfname ), "%s.tmp", fname );
  if ( c < 0 || c > (int)sizeof(tmpfname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );

  /* get sample unit string */
  if (NULL == XLALUnitAsString( units, sizeof( units ), &series->sampleUnits ))
    XLAL_ERROR( func, XLAL_EFUNC );

  /* construct vector and copy data */
  vect = FrVectNew1D( series->name, FR_VECT_16C, series->data->length, series->deltaT, seconds, units );
  if ( ! vect )
    XLAL_ERROR( func, XLAL_EERR );
  vect->startX[0] = 0.0;
  memcpy( vect->data, series->data->data, series->data->length * sizeof( *series->data->data ) );

  frame = XLALFrameNew( &series->epoch, duration, ifo, run, frnum, detectorFlag );
  if ( ! frame )
  {
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EFUNC );
  }

  proc = FrProcDataNewV( frame, vect );
  if ( ! proc )
  {
    FrameFree( frame );
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EERR );
  }
  FrStrCpy( &proc->comment, comment );
  proc->timeOffset = 0;
  proc->type       = 1;
  proc->subType    = 0;
  proc->tRange     = duration;
  proc->fShift     = series->f0;
  proc->phase      = 0.0;
  proc->BW         = 0.0;

  /* write first to tmpfile then rename it */
  frfile = FrFileONew( tmpfname, 0 );
  if ( ! frfile )
  {
    FrameFree( frame ); /* this frees proc and vect */
    XLAL_ERROR( func, XLAL_EIO );
  }
  if ( FR_OK != FrameWrite( frame, frfile ) )
  {
    FrameFree( frame ); /* this frees proc and vect */
    FrFileFree( frfile );
    XLAL_ERROR( func, XLAL_EERR );
  }

  FrFileOEnd( frfile );
  FrameFree( frame ); /* this frees proc and vect */

  /* now rename tmpfile */
  if ( rename( tmpfname, fname ) < 0 )
    XLAL_ERROR( func, XLAL_ESYS );

  return 0;
}



/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteCOMPLEX16TimeSeries (
    LALStatus		*status,
    COMPLEX16TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  COMPLEX16 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteCOMPLEX16TimeSeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteCOMPLEX16TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_16C, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataD, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}




/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteCOMPLEX16FrequencySeries (
    LALStatus		*status,
    COMPLEX16FrequencySeries 	*series,
    FrOutPar		*params,
    INT4                 subtype
    )
{ /* </lalVerbatim> */
  COMPLEX16 	*data;
  CHAR   hertz[] = "Hz";
  CHAR   comment[] = "Created by LALFrWriteCOMPLEX16FrequencySeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 UNUSED nframes;
  REAL4 deltaT;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteCOMPLEX16TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  deltaT = series->deltaF ? 1.0 / series->deltaF : 1;
  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  /* nframes = params->nframes; */
  nframes = 1; /* params->nframes is ignored */
  {
    UINT4 ncpy;
    struct FrameH     *frame;
    struct FrVect     *vect;
    struct FrProcData *proc;
    ncpy = series->data->length;
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /* NOTE: always make a proc channel! */
    proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect = FrVectNew1D( series->name, FR_VECT_16C, series->data->length,
        series->deltaF, hertz, units );
    if ( ! vect )
    {
      free( proc );

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect->startX[0] = series->f0;

    FrStrCpy( &proc->name, series->name );
    FrStrCpy( &proc->comment, comment );
    proc->next = frame->procData;
    frame->procData = proc;
    proc->classe = FrProcDataDef();
    proc->type = 2;
    proc->data = vect;
    proc->subType = subtype;
    proc->tRange = deltaT;
    proc->fRange = series->data->length * series->deltaF;
    /* frequency shift is in vector */

    memcpy( vect->dataD, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}





























/* <lalVerbatim file="FrameSeriesCP"> */
int XLALFrGetCOMPLEX8FrequencySeries ( COMPLEX8FrequencySeries *series, FrStream *stream )
{ /* </lalVerbatim> */
  static const char func[] = "XLALFrGetCOMPLEX8FrequencySeries";
  struct FrVect	*vect;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8C )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  memcpy( series->data->data, vect->dataF, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetCOMPLEX8FrequencySeries (
    LALStatus		*status,
    COMPLEX8FrequencySeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  struct FrVect	*vect;
  INITSTATUS( status, "LALFrGetCOMPLEX8FrequencySeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( ! series->data, status, FRAMESTREAMH_ENNUL, FRAMESTREAMH_MSGENNUL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8C )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  memcpy( series->data->data, vect->dataF, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  vect=NULL;

  RETURN( status );
}


int XLALFrGetCOMPLEX8TimeSeriesMetadata ( COMPLEX8TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetCOMPLEX8TimeSeriesMetadata";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 noff;
  INT8		 tnow;
  INT8		 tbeg;
  REAL8          rate;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8C )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  FrVectFree(vect);
  return 0;
}


int XLALFrGetCOMPLEX8TimeSeries ( COMPLEX8TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetCOMPLEX8TimeSeries";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  COMPLEX8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8C )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* end here... just metadata was requested */
  {
    FrVectFree(vect);
    return 0;
  }

  /* check to see if data vector is ok */
  if ( ! series->data->data )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data->length )
    XLAL_ERROR( func, XLAL_EBADLEN );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataF );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataF + noff * mult, ncpy*sizeof(*series->data->data) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    if ( XLALFrNext( stream ) < 0 )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EIO );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ENAME ); /* now channel is missing ... */
    }
    if ( vect->type != FR_VECT_8C )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ETYPE ); /* now type is wrong ... */
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataF, ncpy * sizeof( *series->data->data ) );


    FrVectFree(vect);
    vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos],
      stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    if ( XLALFrNext( stream ) < 0 ) {
      stream->mode = keepmode;
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( ! (stream->state & LAL_FR_GAP) )
      stream->epoch = keep;
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
    stream->state |= LAL_FR_GAP;
  /* FIXME: does this need to cause a failure if mode is set to fail on gaps? */

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetCOMPLEX8TimeSeries (
    LALStatus		*status,
    COMPLEX8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  COMPLEX8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  INITSTATUS( status, "LALFrGetCOMPLEX8TimeSeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8C )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    FrVectFree(vect);
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataF );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataF + noff * mult, ncpy * sizeof( *series->data->data ) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;


  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_8C )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataF, ncpy * sizeof( *series->data->data ) );


  FrVectFree(vect);
  vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! (stream->state & LAL_FR_GAP) )
    {
      stream->epoch = keep;
    }
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
  {
    stream->state |= LAL_FR_GAP;
  }

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }

  /* remove this: the error will be reported on the *next* call! */
  /*
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }
  */

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetCOMPLEX8TimeSeriesMetadata (
    LALStatus		*status,
    COMPLEX8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  void *sequence;

  INITSTATUS (status, "LALFrGetCOMPLEX8TimeSeriesMetadata", FRAMESERIESC);
  ATTATCHSTATUSPTR (status);

  ASSERT (series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);
  ASSERT (stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);

  /* save the sequence address, then wipe the series structure */
  sequence = series->data;
  memset (series, 0, sizeof(*series));

  /* call LALFrGetCOMPLEX8TimeSeries to populate the series' metadata */
  LALFrGetCOMPLEX8TimeSeries (status->statusPtr, series, chanin, stream);
  CHECKSTATUSPTR (status);

  /* restore the sequence address */
  series->data = sequence;

  DETATCHSTATUSPTR (status);
  RETURN (status);
}


COMPLEX8TimeSeries *XLALFrReadCOMPLEX8TimeSeries (
	FrStream *stream,
	const char *chname,
	const LIGOTimeGPS *start,
	REAL8 duration,
	size_t lengthlimit
)
{
	static const char func[] = "XLALFrReadCOMPLEX8TimeSeries";
	COMPLEX8TimeSeries *series;
	size_t length;

	/* create and initialize a zero-length time series vector */
	series = XLALCreateCOMPLEX8TimeSeries (chname, start, 0.0, 0.0, &lalADCCountUnit, 0);
	if(!series)
		XLAL_ERROR_NULL (func, XLAL_EFUNC);

	/* get the time series meta-data */
	if(XLALFrGetCOMPLEX8TimeSeriesMetadata (series, stream)) {
		XLALDestroyCOMPLEX8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* resize the time series to the correct number of samples */
	length = duration / series->deltaT;
	if(lengthlimit && (lengthlimit < length))
		length = lengthlimit;
	if(!XLALResizeCOMPLEX8TimeSeries (series, 0, length)) {
		XLALDestroyCOMPLEX8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* read the data */
	if(XLALFrSeek (stream, start) || XLALFrGetCOMPLEX8TimeSeries (series, stream)) {
		XLALDestroyCOMPLEX8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	return(series);
}

























int XLALFrWriteCOMPLEX8TimeSeries ( COMPLEX8TimeSeries *series, int frnum )
{
  static const char func[] = "XLALFrWriteCOMPLEX8TimeSeries";
  char fname[FILENAME_MAX];
  char tmpfname[FILENAME_MAX];
  char comment[] = "XLALFrWriteCOMPLEX8TimeSeries $Id$";
  char seconds[] = "s";
  char units[LALUnitTextSize];
  const int run = 0; /* this routine always sets run number to zero */
  char site = 0;
  char ifo[3] = "\0\0\0";
  char description[FILENAME_MAX] = "";
  char *s;
  int detector;
  int detectorFlag = 0;
  REAL8 duration;
  INT4 t0;
  INT4 dt;
  struct FrProcData *proc;
  struct FrVect *vect;
  struct FrameH *frame;
  struct FrFile *frfile;
  int c;

  if ( ! series )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data || ! series->data->data )
    XLAL_ERROR( func, XLAL_EINVAL );

  /* if channel name is of the form Xn:... then Xn is the ifo name */
  if ( isupper( series->name[0] ) )
  {
    site = series->name[0];
    if ( isdigit( series->name[1] ) && series->name[2] == ':' )
    {
      memcpy( ifo, series->name, 2 );
      /* figure out what detector this is */
      for ( detector = 0; detector < LAL_NUM_DETECTORS; ++detector )
        if ( ! strcmp( ifo, lalCachedDetectors[detector].frDetector.prefix ) )
        {
          detectorFlag = 1 << 2 * detector;
          break;
        }
    }
  }

  /* if site was not set, do something sensible with it */
  if ( ! site )
    site = 'F'; /* fake site */

  /* transform channel name to be a valid description field of file name */
  memset( description, 0, sizeof( description ) );
  strncpy( description, series->name, sizeof( description) - 1 );
  s = description;
  while ( *s )
  {
    if ( ! isalnum( *s ) )
      *s = '_'; /* replace invalid description character with underscore */
    ++s;
  }

  /* compute duration of the data and the stop and end times of frame file */
  duration = series->data->length * series->deltaT;
  t0 = series->epoch.gpsSeconds;
  dt = (INT4)ceil( XLALGPSGetREAL8( &series->epoch ) + duration ) - t0;
  if ( t0 < 0 || dt < 1 )
    XLAL_ERROR( func, XLAL_ETIME );

  /* construct file name */
  c = snprintf( fname, sizeof( fname ), "%c-%s-%d-%d.gwf", site, description, t0, dt );
  if ( c < 0 || c > (int)sizeof(fname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );
  c = snprintf( tmpfname, sizeof( tmpfname ), "%s.tmp", fname );
  if ( c < 0 || c > (int)sizeof(tmpfname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );

  /* get sample unit string */
  if (NULL == XLALUnitAsString( units, sizeof( units ), &series->sampleUnits ))
    XLAL_ERROR( func, XLAL_EFUNC );

  /* construct vector and copy data */
  vect = FrVectNew1D( series->name, FR_VECT_8C, series->data->length, series->deltaT, seconds, units );
  if ( ! vect )
    XLAL_ERROR( func, XLAL_EERR );
  vect->startX[0] = 0.0;
  memcpy( vect->data, series->data->data, series->data->length * sizeof( *series->data->data ) );

  frame = XLALFrameNew( &series->epoch, duration, ifo, run, frnum, detectorFlag );
  if ( ! frame )
  {
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EFUNC );
  }

  proc = FrProcDataNewV( frame, vect );
  if ( ! proc )
  {
    FrameFree( frame );
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EERR );
  }
  FrStrCpy( &proc->comment, comment );
  proc->timeOffset = 0;
  proc->type       = 1;
  proc->subType    = 0;
  proc->tRange     = duration;
  proc->fShift     = series->f0;
  proc->phase      = 0.0;
  proc->BW         = 0.0;

  /* write first to tmpfile then rename it */
  frfile = FrFileONew( tmpfname, 0 );
  if ( ! frfile )
  {
    FrameFree( frame ); /* this frees proc and vect */
    XLAL_ERROR( func, XLAL_EIO );
  }
  if ( FR_OK != FrameWrite( frame, frfile ) )
  {
    FrameFree( frame ); /* this frees proc and vect */
    FrFileFree( frfile );
    XLAL_ERROR( func, XLAL_EERR );
  }

  FrFileOEnd( frfile );
  FrameFree( frame ); /* this frees proc and vect */

  /* now rename tmpfile */
  if ( rename( tmpfname, fname ) < 0 )
    XLAL_ERROR( func, XLAL_ESYS );

  return 0;
}



/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteCOMPLEX8TimeSeries (
    LALStatus		*status,
    COMPLEX8TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  COMPLEX8 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteCOMPLEX8TimeSeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteCOMPLEX8TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_8C, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataD, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}




/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteCOMPLEX8FrequencySeries (
    LALStatus		*status,
    COMPLEX8FrequencySeries 	*series,
    FrOutPar		*params,
    INT4                 subtype
    )
{ /* </lalVerbatim> */
  COMPLEX8 	*data;
  CHAR   hertz[] = "Hz";
  CHAR   comment[] = "Created by LALFrWriteCOMPLEX8FrequencySeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 UNUSED nframes;
  REAL4 deltaT;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteCOMPLEX8TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  deltaT = series->deltaF ? 1.0 / series->deltaF : 1;
  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  /* nframes = params->nframes; */
  nframes = 1; /* params->nframes is ignored */
  {
    UINT4 ncpy;
    struct FrameH     *frame;
    struct FrVect     *vect;
    struct FrProcData *proc;
    ncpy = series->data->length;
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /* NOTE: always make a proc channel! */
    proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect = FrVectNew1D( series->name, FR_VECT_8C, series->data->length,
        series->deltaF, hertz, units );
    if ( ! vect )
    {
      free( proc );

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect->startX[0] = series->f0;

    FrStrCpy( &proc->name, series->name );
    FrStrCpy( &proc->comment, comment );
    proc->next = frame->procData;
    frame->procData = proc;
    proc->classe = FrProcDataDef();
    proc->type = 2;
    proc->data = vect;
    proc->subType = subtype;
    proc->tRange = deltaT;
    proc->fRange = series->data->length * series->deltaF;
    /* frequency shift is in vector */

    memcpy( vect->dataD, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}





























/* <lalVerbatim file="FrameSeriesCP"> */
int XLALFrGetREAL8FrequencySeries ( REAL8FrequencySeries *series, FrStream *stream )
{ /* </lalVerbatim> */
  static const char func[] = "XLALFrGetREAL8FrequencySeries";
  struct FrVect	*vect;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8R )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  memcpy( series->data->data, vect->dataD, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL8FrequencySeries (
    LALStatus		*status,
    REAL8FrequencySeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  struct FrVect	*vect;
  INITSTATUS( status, "LALFrGetREAL8FrequencySeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( ! series->data, status, FRAMESTREAMH_ENNUL, FRAMESTREAMH_MSGENNUL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8R )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  memcpy( series->data->data, vect->dataD, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  vect=NULL;

  RETURN( status );
}


int XLALFrGetREAL8TimeSeriesMetadata ( REAL8TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetREAL8TimeSeriesMetadata";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 noff;
  INT8		 tnow;
  INT8		 tbeg;
  REAL8          rate;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8R )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  FrVectFree(vect);
  return 0;
}


int XLALFrGetREAL8TimeSeries ( REAL8TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetREAL8TimeSeries";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  REAL8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8R )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* end here... just metadata was requested */
  {
    FrVectFree(vect);
    return 0;
  }

  /* check to see if data vector is ok */
  if ( ! series->data->data )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data->length )
    XLAL_ERROR( func, XLAL_EBADLEN );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataD );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataD + noff * mult, ncpy*sizeof(*series->data->data) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    if ( XLALFrNext( stream ) < 0 )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EIO );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ENAME ); /* now channel is missing ... */
    }
    if ( vect->type != FR_VECT_8R )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ETYPE ); /* now type is wrong ... */
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataD, ncpy * sizeof( *series->data->data ) );


    FrVectFree(vect);
    vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos],
      stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    if ( XLALFrNext( stream ) < 0 ) {
      stream->mode = keepmode;
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( ! (stream->state & LAL_FR_GAP) )
      stream->epoch = keep;
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
    stream->state |= LAL_FR_GAP;
  /* FIXME: does this need to cause a failure if mode is set to fail on gaps? */

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL8TimeSeries (
    LALStatus		*status,
    REAL8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  REAL8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  INITSTATUS( status, "LALFrGetREAL8TimeSeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8R )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    FrVectFree(vect);
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataD );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataD + noff * mult, ncpy * sizeof( *series->data->data ) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;


  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_8R )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataD, ncpy * sizeof( *series->data->data ) );


  FrVectFree(vect);
  vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! (stream->state & LAL_FR_GAP) )
    {
      stream->epoch = keep;
    }
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
  {
    stream->state |= LAL_FR_GAP;
  }

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }

  /* remove this: the error will be reported on the *next* call! */
  /*
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }
  */

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL8TimeSeriesMetadata (
    LALStatus		*status,
    REAL8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  void *sequence;

  INITSTATUS (status, "LALFrGetREAL8TimeSeriesMetadata", FRAMESERIESC);
  ATTATCHSTATUSPTR (status);

  ASSERT (series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);
  ASSERT (stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);

  /* save the sequence address, then wipe the series structure */
  sequence = series->data;
  memset (series, 0, sizeof(*series));

  /* call LALFrGetREAL8TimeSeries to populate the series' metadata */
  LALFrGetREAL8TimeSeries (status->statusPtr, series, chanin, stream);
  CHECKSTATUSPTR (status);

  /* restore the sequence address */
  series->data = sequence;

  DETATCHSTATUSPTR (status);
  RETURN (status);
}


REAL8TimeSeries *XLALFrReadREAL8TimeSeries (
	FrStream *stream,
	const char *chname,
	const LIGOTimeGPS *start,
	REAL8 duration,
	size_t lengthlimit
)
{
	static const char func[] = "XLALFrReadREAL8TimeSeries";
	REAL8TimeSeries *series;
	size_t length;

	/* create and initialize a zero-length time series vector */
	series = XLALCreateREAL8TimeSeries (chname, start, 0.0, 0.0, &lalADCCountUnit, 0);
	if(!series)
		XLAL_ERROR_NULL (func, XLAL_EFUNC);

	/* get the time series meta-data */
	if(XLALFrGetREAL8TimeSeriesMetadata (series, stream)) {
		XLALDestroyREAL8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* resize the time series to the correct number of samples */
	length = duration / series->deltaT;
	if(lengthlimit && (lengthlimit < length))
		length = lengthlimit;
	if(!XLALResizeREAL8TimeSeries (series, 0, length)) {
		XLALDestroyREAL8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* read the data */
	if(XLALFrSeek (stream, start) || XLALFrGetREAL8TimeSeries (series, stream)) {
		XLALDestroyREAL8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	return(series);
}

























int XLALFrWriteREAL8TimeSeries ( REAL8TimeSeries *series, int frnum )
{
  static const char func[] = "XLALFrWriteREAL8TimeSeries";
  char fname[FILENAME_MAX];
  char tmpfname[FILENAME_MAX];
  char comment[] = "XLALFrWriteREAL8TimeSeries $Id$";
  char seconds[] = "s";
  char units[LALUnitTextSize];
  const int run = 0; /* this routine always sets run number to zero */
  char site = 0;
  char ifo[3] = "\0\0\0";
  char description[FILENAME_MAX] = "";
  char *s;
  int detector;
  int detectorFlag = 0;
  REAL8 duration;
  INT4 t0;
  INT4 dt;
  struct FrProcData *proc;
  struct FrVect *vect;
  struct FrameH *frame;
  struct FrFile *frfile;
  int c;

  if ( ! series )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data || ! series->data->data )
    XLAL_ERROR( func, XLAL_EINVAL );

  /* if channel name is of the form Xn:... then Xn is the ifo name */
  if ( isupper( series->name[0] ) )
  {
    site = series->name[0];
    if ( isdigit( series->name[1] ) && series->name[2] == ':' )
    {
      memcpy( ifo, series->name, 2 );
      /* figure out what detector this is */
      for ( detector = 0; detector < LAL_NUM_DETECTORS; ++detector )
        if ( ! strcmp( ifo, lalCachedDetectors[detector].frDetector.prefix ) )
        {
          detectorFlag = 1 << 2 * detector;
          break;
        }
    }
  }

  /* if site was not set, do something sensible with it */
  if ( ! site )
    site = 'F'; /* fake site */

  /* transform channel name to be a valid description field of file name */
  memset( description, 0, sizeof( description ) );
  strncpy( description, series->name, sizeof( description) - 1 );
  s = description;
  while ( *s )
  {
    if ( ! isalnum( *s ) )
      *s = '_'; /* replace invalid description character with underscore */
    ++s;
  }

  /* compute duration of the data and the stop and end times of frame file */
  duration = series->data->length * series->deltaT;
  t0 = series->epoch.gpsSeconds;
  dt = (INT4)ceil( XLALGPSGetREAL8( &series->epoch ) + duration ) - t0;
  if ( t0 < 0 || dt < 1 )
    XLAL_ERROR( func, XLAL_ETIME );

  /* construct file name */
  c = snprintf( fname, sizeof( fname ), "%c-%s-%d-%d.gwf", site, description, t0, dt );
  if ( c < 0 || c > (int)sizeof(fname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );
  c = snprintf( tmpfname, sizeof( tmpfname ), "%s.tmp", fname );
  if ( c < 0 || c > (int)sizeof(tmpfname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );

  /* get sample unit string */
  if (NULL == XLALUnitAsString( units, sizeof( units ), &series->sampleUnits ))
    XLAL_ERROR( func, XLAL_EFUNC );

  /* construct vector and copy data */
  vect = FrVectNew1D( series->name, FR_VECT_8R, series->data->length, series->deltaT, seconds, units );
  if ( ! vect )
    XLAL_ERROR( func, XLAL_EERR );
  vect->startX[0] = 0.0;
  memcpy( vect->data, series->data->data, series->data->length * sizeof( *series->data->data ) );

  frame = XLALFrameNew( &series->epoch, duration, ifo, run, frnum, detectorFlag );
  if ( ! frame )
  {
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EFUNC );
  }

  proc = FrProcDataNewV( frame, vect );
  if ( ! proc )
  {
    FrameFree( frame );
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EERR );
  }
  FrStrCpy( &proc->comment, comment );
  proc->timeOffset = 0;
  proc->type       = 1;
  proc->subType    = 0;
  proc->tRange     = duration;
  proc->fShift     = series->f0;
  proc->phase      = 0.0;
  proc->BW         = 0.0;

  /* write first to tmpfile then rename it */
  frfile = FrFileONew( tmpfname, 0 );
  if ( ! frfile )
  {
    FrameFree( frame ); /* this frees proc and vect */
    XLAL_ERROR( func, XLAL_EIO );
  }
  if ( FR_OK != FrameWrite( frame, frfile ) )
  {
    FrameFree( frame ); /* this frees proc and vect */
    FrFileFree( frfile );
    XLAL_ERROR( func, XLAL_EERR );
  }

  FrFileOEnd( frfile );
  FrameFree( frame ); /* this frees proc and vect */

  /* now rename tmpfile */
  if ( rename( tmpfname, fname ) < 0 )
    XLAL_ERROR( func, XLAL_ESYS );

  return 0;
}



/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteREAL8TimeSeries (
    LALStatus		*status,
    REAL8TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  REAL8 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteREAL8TimeSeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteREAL8TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_8R, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataD, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}




/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteREAL8FrequencySeries (
    LALStatus		*status,
    REAL8FrequencySeries 	*series,
    FrOutPar		*params,
    INT4                 subtype
    )
{ /* </lalVerbatim> */
  REAL8 	*data;
  CHAR   hertz[] = "Hz";
  CHAR   comment[] = "Created by LALFrWriteREAL8FrequencySeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 UNUSED nframes;
  REAL4 deltaT;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteREAL8TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  deltaT = series->deltaF ? 1.0 / series->deltaF : 1;
  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  /* nframes = params->nframes; */
  nframes = 1; /* params->nframes is ignored */
  {
    UINT4 ncpy;
    struct FrameH     *frame;
    struct FrVect     *vect;
    struct FrProcData *proc;
    ncpy = series->data->length;
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /* NOTE: always make a proc channel! */
    proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect = FrVectNew1D( series->name, FR_VECT_8R, series->data->length,
        series->deltaF, hertz, units );
    if ( ! vect )
    {
      free( proc );

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect->startX[0] = series->f0;

    FrStrCpy( &proc->name, series->name );
    FrStrCpy( &proc->comment, comment );
    proc->next = frame->procData;
    frame->procData = proc;
    proc->classe = FrProcDataDef();
    proc->type = 2;
    proc->data = vect;
    proc->subType = subtype;
    proc->tRange = deltaT;
    proc->fRange = series->data->length * series->deltaF;
    /* frequency shift is in vector */

    memcpy( vect->dataD, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}





























/* <lalVerbatim file="FrameSeriesCP"> */
int XLALFrGetREAL4FrequencySeries ( REAL4FrequencySeries *series, FrStream *stream )
{ /* </lalVerbatim> */
  static const char func[] = "XLALFrGetREAL4FrequencySeries";
  struct FrVect	*vect;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_4R )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  memcpy( series->data->data, vect->dataF, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL4FrequencySeries (
    LALStatus		*status,
    REAL4FrequencySeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  struct FrVect	*vect;
  INITSTATUS( status, "LALFrGetREAL4FrequencySeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( ! series->data, status, FRAMESTREAMH_ENNUL, FRAMESTREAMH_MSGENNUL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_4R )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  memcpy( series->data->data, vect->dataF, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  vect=NULL;

  RETURN( status );
}


int XLALFrGetREAL4TimeSeriesMetadata ( REAL4TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetREAL4TimeSeriesMetadata";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 noff;
  INT8		 tnow;
  INT8		 tbeg;
  REAL8          rate;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_4R )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  FrVectFree(vect);
  return 0;
}


int XLALFrGetREAL4TimeSeries ( REAL4TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetREAL4TimeSeries";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  REAL4 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_4R )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* end here... just metadata was requested */
  {
    FrVectFree(vect);
    return 0;
  }

  /* check to see if data vector is ok */
  if ( ! series->data->data )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data->length )
    XLAL_ERROR( func, XLAL_EBADLEN );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataF );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataF + noff * mult, ncpy*sizeof(*series->data->data) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    if ( XLALFrNext( stream ) < 0 )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EIO );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ENAME ); /* now channel is missing ... */
    }
    if ( vect->type != FR_VECT_4R )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ETYPE ); /* now type is wrong ... */
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataF, ncpy * sizeof( *series->data->data ) );


    FrVectFree(vect);
    vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos],
      stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    if ( XLALFrNext( stream ) < 0 ) {
      stream->mode = keepmode;
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( ! (stream->state & LAL_FR_GAP) )
      stream->epoch = keep;
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
    stream->state |= LAL_FR_GAP;
  /* FIXME: does this need to cause a failure if mode is set to fail on gaps? */

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL4TimeSeries (
    LALStatus		*status,
    REAL4TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  REAL4 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  INITSTATUS( status, "LALFrGetREAL4TimeSeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_4R )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    FrVectFree(vect);
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataF );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataF + noff * mult, ncpy * sizeof( *series->data->data ) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;


  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_4R )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataF, ncpy * sizeof( *series->data->data ) );


  FrVectFree(vect);
  vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! (stream->state & LAL_FR_GAP) )
    {
      stream->epoch = keep;
    }
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
  {
    stream->state |= LAL_FR_GAP;
  }

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }

  /* remove this: the error will be reported on the *next* call! */
  /*
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }
  */

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetREAL4TimeSeriesMetadata (
    LALStatus		*status,
    REAL4TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  void *sequence;

  INITSTATUS (status, "LALFrGetREAL4TimeSeriesMetadata", FRAMESERIESC);
  ATTATCHSTATUSPTR (status);

  ASSERT (series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);
  ASSERT (stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);

  /* save the sequence address, then wipe the series structure */
  sequence = series->data;
  memset (series, 0, sizeof(*series));

  /* call LALFrGetREAL4TimeSeries to populate the series' metadata */
  LALFrGetREAL4TimeSeries (status->statusPtr, series, chanin, stream);
  CHECKSTATUSPTR (status);

  /* restore the sequence address */
  series->data = sequence;

  DETATCHSTATUSPTR (status);
  RETURN (status);
}


REAL4TimeSeries *XLALFrReadREAL4TimeSeries (
	FrStream *stream,
	const char *chname,
	const LIGOTimeGPS *start,
	REAL8 duration,
	size_t lengthlimit
)
{
	static const char func[] = "XLALFrReadREAL4TimeSeries";
	REAL4TimeSeries *series;
	size_t length;

	/* create and initialize a zero-length time series vector */
	series = XLALCreateREAL4TimeSeries (chname, start, 0.0, 0.0, &lalADCCountUnit, 0);
	if(!series)
		XLAL_ERROR_NULL (func, XLAL_EFUNC);

	/* get the time series meta-data */
	if(XLALFrGetREAL4TimeSeriesMetadata (series, stream)) {
		XLALDestroyREAL4TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* resize the time series to the correct number of samples */
	length = duration / series->deltaT;
	if(lengthlimit && (lengthlimit < length))
		length = lengthlimit;
	if(!XLALResizeREAL4TimeSeries (series, 0, length)) {
		XLALDestroyREAL4TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* read the data */
	if(XLALFrSeek (stream, start) || XLALFrGetREAL4TimeSeries (series, stream)) {
		XLALDestroyREAL4TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	return(series);
}

























int XLALFrWriteREAL4TimeSeries ( REAL4TimeSeries *series, int frnum )
{
  static const char func[] = "XLALFrWriteREAL4TimeSeries";
  char fname[FILENAME_MAX];
  char tmpfname[FILENAME_MAX];
  char comment[] = "XLALFrWriteREAL4TimeSeries $Id$";
  char seconds[] = "s";
  char units[LALUnitTextSize];
  const int run = 0; /* this routine always sets run number to zero */
  char site = 0;
  char ifo[3] = "\0\0\0";
  char description[FILENAME_MAX] = "";
  char *s;
  int detector;
  int detectorFlag = 0;
  REAL8 duration;
  INT4 t0;
  INT4 dt;
  struct FrProcData *proc;
  struct FrVect *vect;
  struct FrameH *frame;
  struct FrFile *frfile;
  int c;

  if ( ! series )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data || ! series->data->data )
    XLAL_ERROR( func, XLAL_EINVAL );

  /* if channel name is of the form Xn:... then Xn is the ifo name */
  if ( isupper( series->name[0] ) )
  {
    site = series->name[0];
    if ( isdigit( series->name[1] ) && series->name[2] == ':' )
    {
      memcpy( ifo, series->name, 2 );
      /* figure out what detector this is */
      for ( detector = 0; detector < LAL_NUM_DETECTORS; ++detector )
        if ( ! strcmp( ifo, lalCachedDetectors[detector].frDetector.prefix ) )
        {
          detectorFlag = 1 << 2 * detector;
          break;
        }
    }
  }

  /* if site was not set, do something sensible with it */
  if ( ! site )
    site = 'F'; /* fake site */

  /* transform channel name to be a valid description field of file name */
  memset( description, 0, sizeof( description ) );
  strncpy( description, series->name, sizeof( description) - 1 );
  s = description;
  while ( *s )
  {
    if ( ! isalnum( *s ) )
      *s = '_'; /* replace invalid description character with underscore */
    ++s;
  }

  /* compute duration of the data and the stop and end times of frame file */
  duration = series->data->length * series->deltaT;
  t0 = series->epoch.gpsSeconds;
  dt = (INT4)ceil( XLALGPSGetREAL8( &series->epoch ) + duration ) - t0;
  if ( t0 < 0 || dt < 1 )
    XLAL_ERROR( func, XLAL_ETIME );

  /* construct file name */
  c = snprintf( fname, sizeof( fname ), "%c-%s-%d-%d.gwf", site, description, t0, dt );
  if ( c < 0 || c > (int)sizeof(fname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );
  c = snprintf( tmpfname, sizeof( tmpfname ), "%s.tmp", fname );
  if ( c < 0 || c > (int)sizeof(tmpfname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );

  /* get sample unit string */
  if (NULL == XLALUnitAsString( units, sizeof( units ), &series->sampleUnits ))
    XLAL_ERROR( func, XLAL_EFUNC );

  /* construct vector and copy data */
  vect = FrVectNew1D( series->name, FR_VECT_4R, series->data->length, series->deltaT, seconds, units );
  if ( ! vect )
    XLAL_ERROR( func, XLAL_EERR );
  vect->startX[0] = 0.0;
  memcpy( vect->data, series->data->data, series->data->length * sizeof( *series->data->data ) );

  frame = XLALFrameNew( &series->epoch, duration, ifo, run, frnum, detectorFlag );
  if ( ! frame )
  {
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EFUNC );
  }

  proc = FrProcDataNewV( frame, vect );
  if ( ! proc )
  {
    FrameFree( frame );
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EERR );
  }
  FrStrCpy( &proc->comment, comment );
  proc->timeOffset = 0;
  proc->type       = 1;
  proc->subType    = 0;
  proc->tRange     = duration;
  proc->fShift     = series->f0;
  proc->phase      = 0.0;
  proc->BW         = 0.0;

  /* write first to tmpfile then rename it */
  frfile = FrFileONew( tmpfname, 0 );
  if ( ! frfile )
  {
    FrameFree( frame ); /* this frees proc and vect */
    XLAL_ERROR( func, XLAL_EIO );
  }
  if ( FR_OK != FrameWrite( frame, frfile ) )
  {
    FrameFree( frame ); /* this frees proc and vect */
    FrFileFree( frfile );
    XLAL_ERROR( func, XLAL_EERR );
  }

  FrFileOEnd( frfile );
  FrameFree( frame ); /* this frees proc and vect */

  /* now rename tmpfile */
  if ( rename( tmpfname, fname ) < 0 )
    XLAL_ERROR( func, XLAL_ESYS );

  return 0;
}



/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteREAL4TimeSeries (
    LALStatus		*status,
    REAL4TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  REAL4 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteREAL4TimeSeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteREAL4TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_4R, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataF, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}




/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteREAL4FrequencySeries (
    LALStatus		*status,
    REAL4FrequencySeries 	*series,
    FrOutPar		*params,
    INT4                 subtype
    )
{ /* </lalVerbatim> */
  REAL4 	*data;
  CHAR   hertz[] = "Hz";
  CHAR   comment[] = "Created by LALFrWriteREAL4FrequencySeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 UNUSED nframes;
  REAL4 deltaT;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteREAL4TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  deltaT = series->deltaF ? 1.0 / series->deltaF : 1;
  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  /* nframes = params->nframes; */
  nframes = 1; /* params->nframes is ignored */
  {
    UINT4 ncpy;
    struct FrameH     *frame;
    struct FrVect     *vect;
    struct FrProcData *proc;
    ncpy = series->data->length;
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /* NOTE: always make a proc channel! */
    proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect = FrVectNew1D( series->name, FR_VECT_4R, series->data->length,
        series->deltaF, hertz, units );
    if ( ! vect )
    {
      free( proc );

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect->startX[0] = series->f0;

    FrStrCpy( &proc->name, series->name );
    FrStrCpy( &proc->comment, comment );
    proc->next = frame->procData;
    frame->procData = proc;
    proc->classe = FrProcDataDef();
    proc->type = 2;
    proc->data = vect;
    proc->subType = subtype;
    proc->tRange = deltaT;
    proc->fRange = series->data->length * series->deltaF;
    /* frequency shift is in vector */

    memcpy( vect->dataF, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}





























/* <lalVerbatim file="FrameSeriesCP"> */
int XLALFrGetINT8FrequencySeries ( INT8FrequencySeries *series, FrStream *stream )
{ /* </lalVerbatim> */
  static const char func[] = "XLALFrGetINT8FrequencySeries";
  struct FrVect	*vect;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  memcpy( series->data->data, vect->dataL, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT8FrequencySeries (
    LALStatus		*status,
    INT8FrequencySeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  struct FrVect	*vect;
  INITSTATUS( status, "LALFrGetINT8FrequencySeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( ! series->data, status, FRAMESTREAMH_ENNUL, FRAMESTREAMH_MSGENNUL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8S )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  memcpy( series->data->data, vect->dataL, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  vect=NULL;

  RETURN( status );
}


int XLALFrGetINT8TimeSeriesMetadata ( INT8TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetINT8TimeSeriesMetadata";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 noff;
  INT8		 tnow;
  INT8		 tbeg;
  REAL8          rate;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  FrVectFree(vect);
  return 0;
}


int XLALFrGetINT8TimeSeries ( INT8TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetINT8TimeSeries";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  INT8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_8S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* end here... just metadata was requested */
  {
    FrVectFree(vect);
    return 0;
  }

  /* check to see if data vector is ok */
  if ( ! series->data->data )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data->length )
    XLAL_ERROR( func, XLAL_EBADLEN );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataL );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataL + noff * mult, ncpy*sizeof(*series->data->data) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    if ( XLALFrNext( stream ) < 0 )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EIO );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ENAME ); /* now channel is missing ... */
    }
    if ( vect->type != FR_VECT_8S )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ETYPE ); /* now type is wrong ... */
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataL, ncpy * sizeof( *series->data->data ) );


    FrVectFree(vect);
    vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos],
      stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    if ( XLALFrNext( stream ) < 0 ) {
      stream->mode = keepmode;
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( ! (stream->state & LAL_FR_GAP) )
      stream->epoch = keep;
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
    stream->state |= LAL_FR_GAP;
  /* FIXME: does this need to cause a failure if mode is set to fail on gaps? */

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT8TimeSeries (
    LALStatus		*status,
    INT8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  INT8 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  INITSTATUS( status, "LALFrGetINT8TimeSeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_8S )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    FrVectFree(vect);
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataL );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataL + noff * mult, ncpy * sizeof( *series->data->data ) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;


  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_8S )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataL, ncpy * sizeof( *series->data->data ) );


  FrVectFree(vect);
  vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! (stream->state & LAL_FR_GAP) )
    {
      stream->epoch = keep;
    }
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
  {
    stream->state |= LAL_FR_GAP;
  }

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }

  /* remove this: the error will be reported on the *next* call! */
  /*
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }
  */

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT8TimeSeriesMetadata (
    LALStatus		*status,
    INT8TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  void *sequence;

  INITSTATUS (status, "LALFrGetINT8TimeSeriesMetadata", FRAMESERIESC);
  ATTATCHSTATUSPTR (status);

  ASSERT (series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);
  ASSERT (stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);

  /* save the sequence address, then wipe the series structure */
  sequence = series->data;
  memset (series, 0, sizeof(*series));

  /* call LALFrGetINT8TimeSeries to populate the series' metadata */
  LALFrGetINT8TimeSeries (status->statusPtr, series, chanin, stream);
  CHECKSTATUSPTR (status);

  /* restore the sequence address */
  series->data = sequence;

  DETATCHSTATUSPTR (status);
  RETURN (status);
}


INT8TimeSeries *XLALFrReadINT8TimeSeries (
	FrStream *stream,
	const char *chname,
	const LIGOTimeGPS *start,
	REAL8 duration,
	size_t lengthlimit
)
{
	static const char func[] = "XLALFrReadINT8TimeSeries";
	INT8TimeSeries *series;
	size_t length;

	/* create and initialize a zero-length time series vector */
	series = XLALCreateINT8TimeSeries (chname, start, 0.0, 0.0, &lalADCCountUnit, 0);
	if(!series)
		XLAL_ERROR_NULL (func, XLAL_EFUNC);

	/* get the time series meta-data */
	if(XLALFrGetINT8TimeSeriesMetadata (series, stream)) {
		XLALDestroyINT8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* resize the time series to the correct number of samples */
	length = duration / series->deltaT;
	if(lengthlimit && (lengthlimit < length))
		length = lengthlimit;
	if(!XLALResizeINT8TimeSeries (series, 0, length)) {
		XLALDestroyINT8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* read the data */
	if(XLALFrSeek (stream, start) || XLALFrGetINT8TimeSeries (series, stream)) {
		XLALDestroyINT8TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	return(series);
}

























int XLALFrWriteINT8TimeSeries ( INT8TimeSeries *series, int frnum )
{
  static const char func[] = "XLALFrWriteINT8TimeSeries";
  char fname[FILENAME_MAX];
  char tmpfname[FILENAME_MAX];
  char comment[] = "XLALFrWriteINT8TimeSeries $Id$";
  char seconds[] = "s";
  char units[LALUnitTextSize];
  const int run = 0; /* this routine always sets run number to zero */
  char site = 0;
  char ifo[3] = "\0\0\0";
  char description[FILENAME_MAX] = "";
  char *s;
  int detector;
  int detectorFlag = 0;
  REAL8 duration;
  INT4 t0;
  INT4 dt;
  struct FrProcData *proc;
  struct FrVect *vect;
  struct FrameH *frame;
  struct FrFile *frfile;
  int c;

  if ( ! series )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data || ! series->data->data )
    XLAL_ERROR( func, XLAL_EINVAL );

  /* if channel name is of the form Xn:... then Xn is the ifo name */
  if ( isupper( series->name[0] ) )
  {
    site = series->name[0];
    if ( isdigit( series->name[1] ) && series->name[2] == ':' )
    {
      memcpy( ifo, series->name, 2 );
      /* figure out what detector this is */
      for ( detector = 0; detector < LAL_NUM_DETECTORS; ++detector )
        if ( ! strcmp( ifo, lalCachedDetectors[detector].frDetector.prefix ) )
        {
          detectorFlag = 1 << 2 * detector;
          break;
        }
    }
  }

  /* if site was not set, do something sensible with it */
  if ( ! site )
    site = 'F'; /* fake site */

  /* transform channel name to be a valid description field of file name */
  memset( description, 0, sizeof( description ) );
  strncpy( description, series->name, sizeof( description) - 1 );
  s = description;
  while ( *s )
  {
    if ( ! isalnum( *s ) )
      *s = '_'; /* replace invalid description character with underscore */
    ++s;
  }

  /* compute duration of the data and the stop and end times of frame file */
  duration = series->data->length * series->deltaT;
  t0 = series->epoch.gpsSeconds;
  dt = (INT4)ceil( XLALGPSGetREAL8( &series->epoch ) + duration ) - t0;
  if ( t0 < 0 || dt < 1 )
    XLAL_ERROR( func, XLAL_ETIME );

  /* construct file name */
  c = snprintf( fname, sizeof( fname ), "%c-%s-%d-%d.gwf", site, description, t0, dt );
  if ( c < 0 || c > (int)sizeof(fname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );
  c = snprintf( tmpfname, sizeof( tmpfname ), "%s.tmp", fname );
  if ( c < 0 || c > (int)sizeof(tmpfname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );

  /* get sample unit string */
  if (NULL == XLALUnitAsString( units, sizeof( units ), &series->sampleUnits ))
    XLAL_ERROR( func, XLAL_EFUNC );

  /* construct vector and copy data */
  vect = FrVectNew1D( series->name, FR_VECT_8S, series->data->length, series->deltaT, seconds, units );
  if ( ! vect )
    XLAL_ERROR( func, XLAL_EERR );
  vect->startX[0] = 0.0;
  memcpy( vect->data, series->data->data, series->data->length * sizeof( *series->data->data ) );

  frame = XLALFrameNew( &series->epoch, duration, ifo, run, frnum, detectorFlag );
  if ( ! frame )
  {
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EFUNC );
  }

  proc = FrProcDataNewV( frame, vect );
  if ( ! proc )
  {
    FrameFree( frame );
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EERR );
  }
  FrStrCpy( &proc->comment, comment );
  proc->timeOffset = 0;
  proc->type       = 1;
  proc->subType    = 0;
  proc->tRange     = duration;
  proc->fShift     = series->f0;
  proc->phase      = 0.0;
  proc->BW         = 0.0;

  /* write first to tmpfile then rename it */
  frfile = FrFileONew( tmpfname, 0 );
  if ( ! frfile )
  {
    FrameFree( frame ); /* this frees proc and vect */
    XLAL_ERROR( func, XLAL_EIO );
  }
  if ( FR_OK != FrameWrite( frame, frfile ) )
  {
    FrameFree( frame ); /* this frees proc and vect */
    FrFileFree( frfile );
    XLAL_ERROR( func, XLAL_EERR );
  }

  FrFileOEnd( frfile );
  FrameFree( frame ); /* this frees proc and vect */

  /* now rename tmpfile */
  if ( rename( tmpfname, fname ) < 0 )
    XLAL_ERROR( func, XLAL_ESYS );

  return 0;
}



/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT8TimeSeries (
    LALStatus		*status,
    INT8TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  INT8 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteINT8TimeSeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT8TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_8S, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataL, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}




/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT8FrequencySeries (
    LALStatus		*status,
    INT8FrequencySeries 	*series,
    FrOutPar		*params,
    INT4                 subtype
    )
{ /* </lalVerbatim> */
  INT8 	*data;
  CHAR   hertz[] = "Hz";
  CHAR   comment[] = "Created by LALFrWriteINT8FrequencySeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 UNUSED nframes;
  REAL4 deltaT;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT8TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  deltaT = series->deltaF ? 1.0 / series->deltaF : 1;
  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  /* nframes = params->nframes; */
  nframes = 1; /* params->nframes is ignored */
  {
    UINT4 ncpy;
    struct FrameH     *frame;
    struct FrVect     *vect;
    struct FrProcData *proc;
    ncpy = series->data->length;
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /* NOTE: always make a proc channel! */
    proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect = FrVectNew1D( series->name, FR_VECT_8S, series->data->length,
        series->deltaF, hertz, units );
    if ( ! vect )
    {
      free( proc );

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect->startX[0] = series->f0;

    FrStrCpy( &proc->name, series->name );
    FrStrCpy( &proc->comment, comment );
    proc->next = frame->procData;
    frame->procData = proc;
    proc->classe = FrProcDataDef();
    proc->type = 2;
    proc->data = vect;
    proc->subType = subtype;
    proc->tRange = deltaT;
    proc->fRange = series->data->length * series->deltaF;
    /* frequency shift is in vector */

    memcpy( vect->dataL, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}





























/* <lalVerbatim file="FrameSeriesCP"> */
int XLALFrGetINT4FrequencySeries ( INT4FrequencySeries *series, FrStream *stream )
{ /* </lalVerbatim> */
  static const char func[] = "XLALFrGetINT4FrequencySeries";
  struct FrVect	*vect;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_4S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  memcpy( series->data->data, vect->dataI, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT4FrequencySeries (
    LALStatus		*status,
    INT4FrequencySeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  struct FrVect	*vect;
  INITSTATUS( status, "LALFrGetINT4FrequencySeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( ! series->data, status, FRAMESTREAMH_ENNUL, FRAMESTREAMH_MSGENNUL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_4S )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  memcpy( series->data->data, vect->dataI, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  vect=NULL;

  RETURN( status );
}


int XLALFrGetINT4TimeSeriesMetadata ( INT4TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetINT4TimeSeriesMetadata";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 noff;
  INT8		 tnow;
  INT8		 tbeg;
  REAL8          rate;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_4S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  FrVectFree(vect);
  return 0;
}


int XLALFrGetINT4TimeSeries ( INT4TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetINT4TimeSeries";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  INT4 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_4S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* end here... just metadata was requested */
  {
    FrVectFree(vect);
    return 0;
  }

  /* check to see if data vector is ok */
  if ( ! series->data->data )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data->length )
    XLAL_ERROR( func, XLAL_EBADLEN );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataI );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataI + noff * mult, ncpy*sizeof(*series->data->data) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    if ( XLALFrNext( stream ) < 0 )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EIO );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ENAME ); /* now channel is missing ... */
    }
    if ( vect->type != FR_VECT_4S )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ETYPE ); /* now type is wrong ... */
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataI, ncpy * sizeof( *series->data->data ) );


    FrVectFree(vect);
    vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos],
      stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    if ( XLALFrNext( stream ) < 0 ) {
      stream->mode = keepmode;
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( ! (stream->state & LAL_FR_GAP) )
      stream->epoch = keep;
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
    stream->state |= LAL_FR_GAP;
  /* FIXME: does this need to cause a failure if mode is set to fail on gaps? */

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT4TimeSeries (
    LALStatus		*status,
    INT4TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  INT4 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  INITSTATUS( status, "LALFrGetINT4TimeSeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_4S )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    FrVectFree(vect);
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataI );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataI + noff * mult, ncpy * sizeof( *series->data->data ) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;


  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_4S )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataI, ncpy * sizeof( *series->data->data ) );


  FrVectFree(vect);
  vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! (stream->state & LAL_FR_GAP) )
    {
      stream->epoch = keep;
    }
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
  {
    stream->state |= LAL_FR_GAP;
  }

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }

  /* remove this: the error will be reported on the *next* call! */
  /*
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }
  */

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT4TimeSeriesMetadata (
    LALStatus		*status,
    INT4TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  void *sequence;

  INITSTATUS (status, "LALFrGetINT4TimeSeriesMetadata", FRAMESERIESC);
  ATTATCHSTATUSPTR (status);

  ASSERT (series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);
  ASSERT (stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);

  /* save the sequence address, then wipe the series structure */
  sequence = series->data;
  memset (series, 0, sizeof(*series));

  /* call LALFrGetINT4TimeSeries to populate the series' metadata */
  LALFrGetINT4TimeSeries (status->statusPtr, series, chanin, stream);
  CHECKSTATUSPTR (status);

  /* restore the sequence address */
  series->data = sequence;

  DETATCHSTATUSPTR (status);
  RETURN (status);
}


INT4TimeSeries *XLALFrReadINT4TimeSeries (
	FrStream *stream,
	const char *chname,
	const LIGOTimeGPS *start,
	REAL8 duration,
	size_t lengthlimit
)
{
	static const char func[] = "XLALFrReadINT4TimeSeries";
	INT4TimeSeries *series;
	size_t length;

	/* create and initialize a zero-length time series vector */
	series = XLALCreateINT4TimeSeries (chname, start, 0.0, 0.0, &lalADCCountUnit, 0);
	if(!series)
		XLAL_ERROR_NULL (func, XLAL_EFUNC);

	/* get the time series meta-data */
	if(XLALFrGetINT4TimeSeriesMetadata (series, stream)) {
		XLALDestroyINT4TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* resize the time series to the correct number of samples */
	length = duration / series->deltaT;
	if(lengthlimit && (lengthlimit < length))
		length = lengthlimit;
	if(!XLALResizeINT4TimeSeries (series, 0, length)) {
		XLALDestroyINT4TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* read the data */
	if(XLALFrSeek (stream, start) || XLALFrGetINT4TimeSeries (series, stream)) {
		XLALDestroyINT4TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	return(series);
}

























int XLALFrWriteINT4TimeSeries ( INT4TimeSeries *series, int frnum )
{
  static const char func[] = "XLALFrWriteINT4TimeSeries";
  char fname[FILENAME_MAX];
  char tmpfname[FILENAME_MAX];
  char comment[] = "XLALFrWriteINT4TimeSeries $Id$";
  char seconds[] = "s";
  char units[LALUnitTextSize];
  const int run = 0; /* this routine always sets run number to zero */
  char site = 0;
  char ifo[3] = "\0\0\0";
  char description[FILENAME_MAX] = "";
  char *s;
  int detector;
  int detectorFlag = 0;
  REAL8 duration;
  INT4 t0;
  INT4 dt;
  struct FrProcData *proc;
  struct FrVect *vect;
  struct FrameH *frame;
  struct FrFile *frfile;
  int c;

  if ( ! series )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data || ! series->data->data )
    XLAL_ERROR( func, XLAL_EINVAL );

  /* if channel name is of the form Xn:... then Xn is the ifo name */
  if ( isupper( series->name[0] ) )
  {
    site = series->name[0];
    if ( isdigit( series->name[1] ) && series->name[2] == ':' )
    {
      memcpy( ifo, series->name, 2 );
      /* figure out what detector this is */
      for ( detector = 0; detector < LAL_NUM_DETECTORS; ++detector )
        if ( ! strcmp( ifo, lalCachedDetectors[detector].frDetector.prefix ) )
        {
          detectorFlag = 1 << 2 * detector;
          break;
        }
    }
  }

  /* if site was not set, do something sensible with it */
  if ( ! site )
    site = 'F'; /* fake site */

  /* transform channel name to be a valid description field of file name */
  memset( description, 0, sizeof( description ) );
  strncpy( description, series->name, sizeof( description) - 1 );
  s = description;
  while ( *s )
  {
    if ( ! isalnum( *s ) )
      *s = '_'; /* replace invalid description character with underscore */
    ++s;
  }

  /* compute duration of the data and the stop and end times of frame file */
  duration = series->data->length * series->deltaT;
  t0 = series->epoch.gpsSeconds;
  dt = (INT4)ceil( XLALGPSGetREAL8( &series->epoch ) + duration ) - t0;
  if ( t0 < 0 || dt < 1 )
    XLAL_ERROR( func, XLAL_ETIME );

  /* construct file name */
  c = snprintf( fname, sizeof( fname ), "%c-%s-%d-%d.gwf", site, description, t0, dt );
  if ( c < 0 || c > (int)sizeof(fname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );
  c = snprintf( tmpfname, sizeof( tmpfname ), "%s.tmp", fname );
  if ( c < 0 || c > (int)sizeof(tmpfname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );

  /* get sample unit string */
  if (NULL == XLALUnitAsString( units, sizeof( units ), &series->sampleUnits ))
    XLAL_ERROR( func, XLAL_EFUNC );

  /* construct vector and copy data */
  vect = FrVectNew1D( series->name, FR_VECT_4S, series->data->length, series->deltaT, seconds, units );
  if ( ! vect )
    XLAL_ERROR( func, XLAL_EERR );
  vect->startX[0] = 0.0;
  memcpy( vect->data, series->data->data, series->data->length * sizeof( *series->data->data ) );

  frame = XLALFrameNew( &series->epoch, duration, ifo, run, frnum, detectorFlag );
  if ( ! frame )
  {
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EFUNC );
  }

  proc = FrProcDataNewV( frame, vect );
  if ( ! proc )
  {
    FrameFree( frame );
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EERR );
  }
  FrStrCpy( &proc->comment, comment );
  proc->timeOffset = 0;
  proc->type       = 1;
  proc->subType    = 0;
  proc->tRange     = duration;
  proc->fShift     = series->f0;
  proc->phase      = 0.0;
  proc->BW         = 0.0;

  /* write first to tmpfile then rename it */
  frfile = FrFileONew( tmpfname, 0 );
  if ( ! frfile )
  {
    FrameFree( frame ); /* this frees proc and vect */
    XLAL_ERROR( func, XLAL_EIO );
  }
  if ( FR_OK != FrameWrite( frame, frfile ) )
  {
    FrameFree( frame ); /* this frees proc and vect */
    FrFileFree( frfile );
    XLAL_ERROR( func, XLAL_EERR );
  }

  FrFileOEnd( frfile );
  FrameFree( frame ); /* this frees proc and vect */

  /* now rename tmpfile */
  if ( rename( tmpfname, fname ) < 0 )
    XLAL_ERROR( func, XLAL_ESYS );

  return 0;
}



/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT4TimeSeries (
    LALStatus		*status,
    INT4TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  INT4 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteINT4TimeSeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT4TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_4S, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataI, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}




/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT4FrequencySeries (
    LALStatus		*status,
    INT4FrequencySeries 	*series,
    FrOutPar		*params,
    INT4                 subtype
    )
{ /* </lalVerbatim> */
  INT4 	*data;
  CHAR   hertz[] = "Hz";
  CHAR   comment[] = "Created by LALFrWriteINT4FrequencySeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 UNUSED nframes;
  REAL4 deltaT;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT4TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  deltaT = series->deltaF ? 1.0 / series->deltaF : 1;
  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  /* nframes = params->nframes; */
  nframes = 1; /* params->nframes is ignored */
  {
    UINT4 ncpy;
    struct FrameH     *frame;
    struct FrVect     *vect;
    struct FrProcData *proc;
    ncpy = series->data->length;
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /* NOTE: always make a proc channel! */
    proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect = FrVectNew1D( series->name, FR_VECT_4S, series->data->length,
        series->deltaF, hertz, units );
    if ( ! vect )
    {
      free( proc );

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect->startX[0] = series->f0;

    FrStrCpy( &proc->name, series->name );
    FrStrCpy( &proc->comment, comment );
    proc->next = frame->procData;
    frame->procData = proc;
    proc->classe = FrProcDataDef();
    proc->type = 2;
    proc->data = vect;
    proc->subType = subtype;
    proc->tRange = deltaT;
    proc->fRange = series->data->length * series->deltaF;
    /* frequency shift is in vector */

    memcpy( vect->dataI, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}





























/* <lalVerbatim file="FrameSeriesCP"> */
int XLALFrGetINT2FrequencySeries ( INT2FrequencySeries *series, FrStream *stream )
{ /* </lalVerbatim> */
  static const char func[] = "XLALFrGetINT2FrequencySeries";
  struct FrVect	*vect;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_2S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ENOMEM );
  }
  memcpy( series->data->data, vect->dataS, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT2FrequencySeries (
    LALStatus		*status,
    INT2FrequencySeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  struct FrVect	*vect;
  INITSTATUS( status, "LALFrGetINT2FrequencySeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( ! series->data, status, FRAMESTREAMH_ENNUL, FRAMESTREAMH_MSGENNUL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_2S )
  {
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

#if defined FR_VERS && FR_VERS >= 5000
  series->epoch.gpsSeconds     = floor( vect->GTime );
  series->epoch.gpsNanoSeconds = floor( 1e9 * ( vect->GTime - floor( vect->GTime ) ) );
#else
  series->epoch.gpsSeconds     = vect->GTimeS;
  series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
  series->deltaF = vect->dx[0];
  series->f0 = 0; /* FIXME: should get correct value... */
  series->sampleUnits = lalADCCountUnit;
  series->data = LALCalloc( 1, sizeof( *series->data ) );
  if ( ! series->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  series->data->length = vect->nData;
  series->data->data = LALMalloc( series->data->length * sizeof( *series->data->data ) );
  if ( ! series->data->data )
  {
    ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
  }
  memcpy( series->data->data, vect->dataS, series->data->length * sizeof( *series->data->data ) );

  FrVectFree(vect);

  vect=NULL;

  RETURN( status );
}


int XLALFrGetINT2TimeSeriesMetadata ( INT2TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetINT2TimeSeriesMetadata";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 noff;
  INT8		 tnow;
  INT8		 tbeg;
  REAL8          rate;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_2S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  FrVectFree(vect);
  return 0;
}


int XLALFrGetINT2TimeSeries ( INT2TimeSeries *series, FrStream *stream )
{
  static const char func[] = "XLALFrGetINT2TimeSeries";
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  INT2 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );
  if ( stream->state & LAL_FR_END )
    XLAL_ERROR( func, XLAL_EIO );

  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
    XLAL_ERROR( func, XLAL_ENAME ); /* couldn't find channel */
  if ( vect->type != FR_VECT_2S )
    XLAL_ERROR( func, XLAL_ETYPE ); /* data has wrong type */

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );

  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* end here... just metadata was requested */
  {
    FrVectFree(vect);
    return 0;
  }

  /* check to see if data vector is ok */
  if ( ! series->data->data )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data->length )
    XLAL_ERROR( func, XLAL_EBADLEN );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataS );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    XLAL_ERROR( func, XLAL_ETIME ); /* invalid time offset */
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataS + noff * mult, ncpy*sizeof(*series->data->data) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;

  /* if still data remaining */
  while ( need )
  {
    if ( XLALFrNext( stream ) < 0 )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_EIO );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ENAME ); /* now channel is missing ... */
    }
    if ( vect->type != FR_VECT_2S )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      XLAL_ERROR( func, XLAL_ETYPE ); /* now type is wrong ... */
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataS, ncpy * sizeof( *series->data->data ) );


    FrVectFree(vect);
    vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos],
      stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    if ( XLALFrNext( stream ) < 0 ) {
      stream->mode = keepmode;
      XLAL_ERROR( func, XLAL_EFUNC );
    }
    if ( ! (stream->state & LAL_FR_GAP) )
      stream->epoch = keep;
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
    stream->state |= LAL_FR_GAP;
  /* FIXME: does this need to cause a failure if mode is set to fail on gaps? */

  if ( stream->state & LAL_FR_ERR )
    XLAL_ERROR( func, XLAL_EIO );

  return 0;
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT2TimeSeries (
    LALStatus		*status,
    INT2TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  const REAL8    fuzz = 0.1 / 16384.0; /* smallest discernable unit of time */
  struct FrVect	*vect;
  UINT4		 need;
  UINT4		 noff;
  UINT4          mult;
  UINT4		 ncpy;
  INT2 		*dest;
  INT8		 tnow;
  INT8		 tbeg;
  INT8           tend;
  REAL8          rate;
  INT4           gap = 0;

  INITSTATUS( status, "LALFrGetINT2TimeSeries", FRAMESERIESC );

  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }

  strncpy( series->name, chanin->name, sizeof( series->name ) );
  vect = loadFrVect( stream, series->name );
  if ( ! vect || ! vect->data )
  {
    ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
  }
  if ( vect->type != FR_VECT_2S )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
  }

  tnow = EPOCH_TO_I8TIME( stream->epoch );
#if defined FR_VERS && FR_VERS >= 5000
  tbeg = 1e9 * vect->GTime;
#else
  tbeg = SECNAN_TO_I8TIME( vect->GTimeS, vect->GTimeN );
#endif
  if ( tnow + 1000 < tbeg )  /* added 1000 ns to account for double precision */
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* compute number of points offset very carefully:
   * if current time is within fuzz of a sample, get that sample;
   * otherwise get the sample just after the requested time */
  rate = vect->dx[0] ? 1.0 / vect->dx[0] : 0.0;
  noff = ceil( ( 1e-9 * ( tnow - tbeg ) - fuzz ) * rate );

  /* adjust current time to be exactly the first sample
   * (rounded to nearest nanosecond) */
  tnow = tbeg + floor( 1e9 * noff * vect->dx[0] + 0.5 );


  SET_EPOCH( &series->epoch, tnow );
  series->deltaT = vect->dx[0];
  series->sampleUnits = lalADCCountUnit;

  if ( ! series->data ) /* no data requested: return now */
  {
    FrVectFree(vect);
    RETURN( status );
  }
  ASSERT( series->data->data, status, FRAMESTREAMH_ENULL,
      FRAMESTREAMH_MSGENULL );
  ASSERT( series->data->length > 0, status, FRAMESTREAMH_ESIZE,
      FRAMESTREAMH_MSGESIZE );

  ATTATCHSTATUSPTR( status );

  /* mult is two if output series is complex */
  mult = sizeof( *series->data->data ) / sizeof( *vect->dataS );
  dest = series->data->data;
  need = series->data->length;
  if ( noff > vect->nData )
  {
    FrVectFree(vect);
    ABORT( status, FRAMESTREAMH_ETIME, FRAMESTREAMH_MSGETIME );
  }

  /* number of points to copy */
  ncpy = ( vect->nData - noff < need ) ? ( vect->nData - noff ) : need;
  memcpy( dest, vect->dataS + noff * mult, ncpy * sizeof( *series->data->data ) );

  FrVectFree(vect);
  vect=NULL;

  dest += ncpy;
  need -= ncpy;


  /* if still data remaining */
  while ( need )
  {
    LALFrNext( status->statusPtr, stream );
    BEGINFAIL( status )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
    }
    ENDFAIL( status );
    if ( stream->state & LAL_FR_END )
    {
      if(vect) FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
    }

    /* load more data */
    vect = loadFrVect( stream, series->name );
    if ( ! vect || ! vect->data )
    {
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ECHAN, FRAMESTREAMH_MSGECHAN );
    }
    if ( vect->type != FR_VECT_2S )
    {
      FrVectFree(vect);
      memset( dest, 0, need * sizeof( *series->data->data ) );
      ABORT( status, FRAMESTREAMH_ETYPE, FRAMESTREAMH_MSGETYPE );
    }

    if ( stream->state & LAL_FR_GAP ) /* gap in data */
    {
      dest = series->data->data;
      need = series->data->length;
#if defined FR_VERS && FR_VERS >= 5000
      series->epoch.gpsSeconds = floor( vect->GTime );
      series->epoch.gpsNanoSeconds = (INT8)( 1e9 * vect->GTime ) % (INT8)1000000000;
#else
      series->epoch.gpsSeconds = vect->GTimeS;
      series->epoch.gpsNanoSeconds = vect->GTimeN;
#endif
      gap = 1; /* need to record this because next FrNext will erase! */
    }

    /* copy data */
    ncpy = vect->nData < need ? vect->nData : need;
    memcpy( dest, vect->dataS, ncpy * sizeof( *series->data->data ) );


  FrVectFree(vect);
  vect=NULL;


    dest += ncpy;
    need -= ncpy;
  }

  /* update stream start time very carefully:
   * start time must be the exact time of the next sample, rounded to the
   * nearest nanosecond */
  SET_EPOCH( &stream->epoch, EPOCH_TO_I8TIME( series->epoch )
      + (INT8)floor( 1e9 * series->data->length * series->deltaT + 0.5 ) );

  /* check to see that we are still within current frame */
  tend  = SECNAN_TO_I8TIME( stream->file->toc->GTimeS[stream->pos], stream->file->toc->GTimeN[stream->pos] );
  tend += (INT8)floor( 1e9 * stream->file->toc->dt[stream->pos] );
  if ( tend <= EPOCH_TO_I8TIME( stream->epoch ) )
  {
    int keepmode = stream->mode;
    LIGOTimeGPS keep;
    keep = stream->epoch;
    /* advance a frame */
    /* failure is benign, so we return results */
    stream->mode |= LAL_FR_IGNOREGAP_MODE;
    TRY( LALFrNext( status->statusPtr, stream ), status );
    if ( ! (stream->state & LAL_FR_GAP) )
    {
      stream->epoch = keep;
    }
    stream->mode = keepmode;
  }

  if ( gap ) /* there was a gap in the data */
  {
    stream->state |= LAL_FR_GAP;
  }

  if ( stream->state & LAL_FR_ERR )
  {
    ABORT( status, FRAMESTREAMH_ERROR, FRAMESTREAMH_MSGERROR );
  }

  /* remove this: the error will be reported on the *next* call! */
  /*
  if ( stream->state & LAL_FR_END )
  {
    ABORT( status, FRAMESTREAMH_EDONE, FRAMESTREAMH_MSGEDONE );
  }
  */

  DETATCHSTATUSPTR( status );
  RETURN( status );
}


/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrGetINT2TimeSeriesMetadata (
    LALStatus		*status,
    INT2TimeSeries	*series,
    FrChanIn		*chanin,
    FrStream		*stream
    )
{ /* </lalVerbatim> */
  void *sequence;

  INITSTATUS (status, "LALFrGetINT2TimeSeriesMetadata", FRAMESERIESC);
  ATTATCHSTATUSPTR (status);

  ASSERT (series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);
  ASSERT (stream, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL);

  /* save the sequence address, then wipe the series structure */
  sequence = series->data;
  memset (series, 0, sizeof(*series));

  /* call LALFrGetINT2TimeSeries to populate the series' metadata */
  LALFrGetINT2TimeSeries (status->statusPtr, series, chanin, stream);
  CHECKSTATUSPTR (status);

  /* restore the sequence address */
  series->data = sequence;

  DETATCHSTATUSPTR (status);
  RETURN (status);
}


INT2TimeSeries *XLALFrReadINT2TimeSeries (
	FrStream *stream,
	const char *chname,
	const LIGOTimeGPS *start,
	REAL8 duration,
	size_t lengthlimit
)
{
	static const char func[] = "XLALFrReadINT2TimeSeries";
	INT2TimeSeries *series;
	size_t length;

	/* create and initialize a zero-length time series vector */
	series = XLALCreateINT2TimeSeries (chname, start, 0.0, 0.0, &lalADCCountUnit, 0);
	if(!series)
		XLAL_ERROR_NULL (func, XLAL_EFUNC);

	/* get the time series meta-data */
	if(XLALFrGetINT2TimeSeriesMetadata (series, stream)) {
		XLALDestroyINT2TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* resize the time series to the correct number of samples */
	length = duration / series->deltaT;
	if(lengthlimit && (lengthlimit < length))
		length = lengthlimit;
	if(!XLALResizeINT2TimeSeries (series, 0, length)) {
		XLALDestroyINT2TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	/* read the data */
	if(XLALFrSeek (stream, start) || XLALFrGetINT2TimeSeries (series, stream)) {
		XLALDestroyINT2TimeSeries (series);
		XLAL_ERROR_NULL (func, XLAL_EFUNC);
	}

	return(series);
}

























int XLALFrWriteINT2TimeSeries ( INT2TimeSeries *series, int frnum )
{
  static const char func[] = "XLALFrWriteINT2TimeSeries";
  char fname[FILENAME_MAX];
  char tmpfname[FILENAME_MAX];
  char comment[] = "XLALFrWriteINT2TimeSeries $Id$";
  char seconds[] = "s";
  char units[LALUnitTextSize];
  const int run = 0; /* this routine always sets run number to zero */
  char site = 0;
  char ifo[3] = "\0\0\0";
  char description[FILENAME_MAX] = "";
  char *s;
  int detector;
  int detectorFlag = 0;
  REAL8 duration;
  INT4 t0;
  INT4 dt;
  struct FrProcData *proc;
  struct FrVect *vect;
  struct FrameH *frame;
  struct FrFile *frfile;
  int c;

  if ( ! series )
    XLAL_ERROR( func, XLAL_EFAULT );
  if ( ! series->data || ! series->data->data )
    XLAL_ERROR( func, XLAL_EINVAL );

  /* if channel name is of the form Xn:... then Xn is the ifo name */
  if ( isupper( series->name[0] ) )
  {
    site = series->name[0];
    if ( isdigit( series->name[1] ) && series->name[2] == ':' )
    {
      memcpy( ifo, series->name, 2 );
      /* figure out what detector this is */
      for ( detector = 0; detector < LAL_NUM_DETECTORS; ++detector )
        if ( ! strcmp( ifo, lalCachedDetectors[detector].frDetector.prefix ) )
        {
          detectorFlag = 1 << 2 * detector;
          break;
        }
    }
  }

  /* if site was not set, do something sensible with it */
  if ( ! site )
    site = 'F'; /* fake site */

  /* transform channel name to be a valid description field of file name */
  memset( description, 0, sizeof( description ) );
  strncpy( description, series->name, sizeof( description) - 1 );
  s = description;
  while ( *s )
  {
    if ( ! isalnum( *s ) )
      *s = '_'; /* replace invalid description character with underscore */
    ++s;
  }

  /* compute duration of the data and the stop and end times of frame file */
  duration = series->data->length * series->deltaT;
  t0 = series->epoch.gpsSeconds;
  dt = (INT4)ceil( XLALGPSGetREAL8( &series->epoch ) + duration ) - t0;
  if ( t0 < 0 || dt < 1 )
    XLAL_ERROR( func, XLAL_ETIME );

  /* construct file name */
  c = snprintf( fname, sizeof( fname ), "%c-%s-%d-%d.gwf", site, description, t0, dt );
  if ( c < 0 || c > (int)sizeof(fname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );
  c = snprintf( tmpfname, sizeof( tmpfname ), "%s.tmp", fname );
  if ( c < 0 || c > (int)sizeof(tmpfname) - 2 )
    XLAL_ERROR( func, XLAL_ENAME );

  /* get sample unit string */
  if (NULL == XLALUnitAsString( units, sizeof( units ), &series->sampleUnits ))
    XLAL_ERROR( func, XLAL_EFUNC );

  /* construct vector and copy data */
  vect = FrVectNew1D( series->name, FR_VECT_2S, series->data->length, series->deltaT, seconds, units );
  if ( ! vect )
    XLAL_ERROR( func, XLAL_EERR );
  vect->startX[0] = 0.0;
  memcpy( vect->data, series->data->data, series->data->length * sizeof( *series->data->data ) );

  frame = XLALFrameNew( &series->epoch, duration, ifo, run, frnum, detectorFlag );
  if ( ! frame )
  {
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EFUNC );
  }

  proc = FrProcDataNewV( frame, vect );
  if ( ! proc )
  {
    FrameFree( frame );
    FrVectFree( vect );
    XLAL_ERROR( func, XLAL_EERR );
  }
  FrStrCpy( &proc->comment, comment );
  proc->timeOffset = 0;
  proc->type       = 1;
  proc->subType    = 0;
  proc->tRange     = duration;
  proc->fShift     = series->f0;
  proc->phase      = 0.0;
  proc->BW         = 0.0;

  /* write first to tmpfile then rename it */
  frfile = FrFileONew( tmpfname, 0 );
  if ( ! frfile )
  {
    FrameFree( frame ); /* this frees proc and vect */
    XLAL_ERROR( func, XLAL_EIO );
  }
  if ( FR_OK != FrameWrite( frame, frfile ) )
  {
    FrameFree( frame ); /* this frees proc and vect */
    FrFileFree( frfile );
    XLAL_ERROR( func, XLAL_EERR );
  }

  FrFileOEnd( frfile );
  FrameFree( frame ); /* this frees proc and vect */

  /* now rename tmpfile */
  if ( rename( tmpfname, fname ) < 0 )
    XLAL_ERROR( func, XLAL_ESYS );

  return 0;
}



/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT2TimeSeries (
    LALStatus		*status,
    INT2TimeSeries 	*series,
    FrOutPar		*params
    )
{ /* </lalVerbatim> */
  INT2 	*data;
  CHAR   seconds[] = "s";
  CHAR   comment[] = "Created by LALFrWriteINT2TimeSeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 nframes;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT2TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * series->data->length * series->deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  nframes = params->nframes;
  while ( nframes-- > 0 )
  {
    UINT4 ncpy;
    struct FrameH *frame;
    struct FrVect *vect;
    if ( nframes )
    {
      ncpy = series->data->length / params->nframes;
    }
    else
    {
      ncpy = series->data->length - ( data - series->data->data );
    }
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = ncpy * series->deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /*** CHECK FSHIFT ***/
    vect = makeFrVect1D( frame, params->type, series->name, comment, seconds,
        units, FR_VECT_2S, series->deltaT ? 1.0 / series->deltaT : 0.0,
        series->f0, series->deltaT, ncpy );
    if ( ! vect )
    {

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    memcpy( vect->dataS, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );
    data += ncpy;
    t += 1e9 * ncpy * series->deltaT;

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}




/* <lalVerbatim file="FrameSeriesCP"> */
void
LALFrWriteINT2FrequencySeries (
    LALStatus		*status,
    INT2FrequencySeries 	*series,
    FrOutPar		*params,
    INT4                 subtype
    )
{ /* </lalVerbatim> */
  INT2 	*data;
  CHAR   hertz[] = "Hz";
  CHAR   comment[] = "Created by LALFrWriteINT2FrequencySeries $Id$";
  CHAR   source[FILENAME_MAX];
  CHAR   fname[FILENAME_MAX];
  CHAR   tmpfname[FILENAME_MAX];
  CHAR   units[LALUnitTextSize];
  CHARVector vnits;
  struct FrFile *frfile;
  UINT4 UNUSED nframes;
  REAL4 deltaT;
  INT8 t;
  INT8 tend;
  INT4 dt;

  INITSTATUS( status, "LALFrWriteINT2TimeSeries", FRAMESERIESC );
  ASSERT( series, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ASSERT( params, status, FRAMESTREAMH_ENULL, FRAMESTREAMH_MSGENULL );
  ATTATCHSTATUSPTR( status );

  vnits.length = sizeof( units );
  vnits.data = units;

  strncpy( source, params->source ? params->source : "F", sizeof( source ) );

  TRY( LALUnitAsString( status->statusPtr, &vnits, &series->sampleUnits ),
      status );

  deltaT = series->deltaF ? 1.0 / series->deltaF : 1;
  t     = EPOCH_TO_I8TIME( series->epoch );
  tend  = t + (INT8)floor( 1e9 * deltaT );
  dt    = tend / (INT8)1000000000;
  dt   -= series->epoch.gpsSeconds;
  dt    = dt < 1 ? 1 : dt; /* must be at least one */

  snprintf( tmpfname, sizeof( tmpfname ), "%s-%s-%d-%d.gwf.tmp", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  snprintf( fname, sizeof( fname ), "%s-%s-%d-%d.gwf", source,
      params->description ? params->description : "UNKNOWN",
      series->epoch.gpsSeconds, dt );
  frfile = FrFileONew( tmpfname, 0 );

  data = series->data->data;
  /* nframes = params->nframes; */
  nframes = 1; /* params->nframes is ignored */
  {
    UINT4 ncpy;
    struct FrameH     *frame;
    struct FrVect     *vect;
    struct FrProcData *proc;
    ncpy = series->data->length;
    frame = FrameHNew( source );
    frame->run = params->run;
    frame->frame = params->frame++;
    frame->GTimeS = t / (INT8)1000000000;
    frame->GTimeN = t % (INT8)1000000000;
    frame->dt = deltaT;
#if !defined FR_VERS || FR_VERS < 5000
    frame->localTime = 0;
#endif

    /* NOTE: always make a proc channel! */
    proc = calloc( 1, sizeof( *proc ) );
    if ( ! proc )
    {
      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect = FrVectNew1D( series->name, FR_VECT_2S, series->data->length,
        series->deltaF, hertz, units );
    if ( ! vect )
    {
      free( proc );

      FrVectFree(vect);
      vect=NULL;

      ABORT( status, FRAMESTREAMH_EALOC, FRAMESTREAMH_MSGEALOC );
    }
    vect->startX[0] = series->f0;

    FrStrCpy( &proc->name, series->name );
    FrStrCpy( &proc->comment, comment );
    proc->next = frame->procData;
    frame->procData = proc;
    proc->classe = FrProcDataDef();
    proc->type = 2;
    proc->data = vect;
    proc->subType = subtype;
    proc->tRange = deltaT;
    proc->fRange = series->data->length * series->deltaF;
    /* frequency shift is in vector */

    memcpy( vect->dataS, data, ncpy * sizeof( *series->data->data ) );

    FrameWrite( frame, frfile );

    FrVectFree(vect);
    vect=NULL;

  }

  FrFileOEnd( frfile );
  rename( tmpfname, fname );

  DETATCHSTATUSPTR( status );
  RETURN( status );
}

